@@ -72,9 +72,9 @@ public:
         mReadingExternalTileset(false)
     {}
 
-    Map *readMap(QIODevice *device, const QString &path);
+    std::unique_ptr<Map> readMap(QIODevice *device, const QString &path);
     SharedTileset readTileset(QIODevice *device, const QString &path);
-    ObjectTemplate *readObjectTemplate(QIODevice *device, const QString &path);
+    std::unique_ptr<ObjectTemplate> readObjectTemplate(QIODevice *device, const QString &path);
 
     bool openFile(QFile *file);
 
@@ -83,9 +83,11 @@ public:
 private:
     void readUnknownElement();
 
-    Map *readMap();
+    std::unique_ptr<Map> readMap();
+    void readMapEditorSettings(Map &map);
 
     SharedTileset readTileset();
+    void readTilesetEditorSettings(Tileset &tileset);
     void readTilesetTile(Tileset &tileset);
     void readTilesetGrid(Tileset &tileset);
     void readTilesetImage(Tileset &tileset);
@@ -93,11 +95,11 @@ private:
     void readTilesetWangSets(Tileset &tileset);
     ImageReference readImage();
 
-    ObjectTemplate *readObjectTemplate();
+    std::unique_ptr<ObjectTemplate> readObjectTemplate();
 
-    Layer *tryReadLayer();
+    std::unique_ptr<Layer> tryReadLayer();
 
-    TileLayer *readTileLayer();
+    std::unique_ptr<TileLayer> readTileLayer();
     void readTileLayerData(TileLayer &tileLayer);
     void readTileLayerRect(TileLayer &tileLayer,
                            Map::LayerDataFormat layerDataFormat,
@@ -121,15 +123,15 @@ private:
      */
     Cell cellForGid(unsigned gid);
 
-    ImageLayer *readImageLayer();
+    std::unique_ptr<ImageLayer> readImageLayer();
     void readImageLayerImage(ImageLayer &imageLayer);
 
-    ObjectGroup *readObjectGroup();
-    MapObject *readObject();
+    std::unique_ptr<ObjectGroup> readObjectGroup();
+    std::unique_ptr<MapObject> readObject();
     QPolygonF readPolygon();
     TextData readObjectText();
 
-    GroupLayer *readGroupLayer();
+    std::unique_ptr<GroupLayer> readGroupLayer();
 
     QVector<Frame> readAnimationFrames();
 
@@ -150,11 +152,11 @@ private:
 } // namespace Internal
 } // namespace Tiled
 
-Map *MapReaderPrivate::readMap(QIODevice *device, const QString &path)
+std::unique_ptr<Map> MapReaderPrivate::readMap(QIODevice *device, const QString &path)
 {
     mError.clear();
     mPath.setPath(path);
-    Map *map = nullptr;
+    std::unique_ptr<Map> map;
 
     xml.setDevice(device);
 
@@ -186,11 +188,11 @@ SharedTileset MapReaderPrivate::readTileset(QIODevice *device, const QString &pa
     return tileset;
 }
 
-ObjectTemplate *MapReaderPrivate::readObjectTemplate(QIODevice *device, const QString &path)
+std::unique_ptr<ObjectTemplate> MapReaderPrivate::readObjectTemplate(QIODevice *device, const QString &path)
 {
     mError.clear();
-    mPath = path;
-    ObjectTemplate *objectTemplate = nullptr;
+    mPath.setPath(path);
+    std::unique_ptr<ObjectTemplate> objectTemplate;
 
     xml.setDevice(device);
 
@@ -235,7 +237,7 @@ void MapReaderPrivate::readUnknownElement()
     xml.skipCurrentElement();
 }
 
-Map *MapReaderPrivate::readMap()
+std::unique_ptr<Map> MapReaderPrivate::readMap()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("map"));
 
@@ -271,15 +273,18 @@ Map *MapReaderPrivate::readMap()
             atts.value(QLatin1String("renderorder")).toString();
     const Map::RenderOrder renderOrder =
             renderOrderFromString(renderOrderString);
+    const QString compressionLevelString =
+            atts.value(QLatin1String("compressionlevel")).toString();
 
     const int nextLayerId = atts.value(QLatin1String("nextlayerid")).toInt();
     const int nextObjectId = atts.value(QLatin1String("nextobjectid")).toInt();
 
-    mMap.reset(new Map(orientation, mapWidth, mapHeight, tileWidth, tileHeight, infinite));
+    mMap = std::make_unique<Map>(orientation, mapWidth, mapHeight, tileWidth, tileHeight, infinite);
     mMap->setHexSideLength(hexSideLength);
     mMap->setStaggerAxis(staggerAxis);
     mMap->setStaggerIndex(staggerIndex);
     mMap->setRenderOrder(renderOrder);
+    mMap->setCompressionLevel(compressionLevelString.toUInt());
     if (nextLayerId)
         mMap->setNextLayerId(nextLayerId);
     if (nextObjectId)
@@ -290,8 +295,10 @@ Map *MapReaderPrivate::readMap()
         mMap->setBackgroundColor(QColor(bgColorString.toString()));
 
     while (xml.readNextStartElement()) {
-        if (Layer *layer = tryReadLayer())
-            mMap->addLayer(layer);
+        if (xml.name() == QLatin1String("editorsettings"))
+            readMapEditorSettings(*mMap);
+        else if (std::unique_ptr<Layer> layer = tryReadLayer())
+            mMap->addLayer(std::move(layer));
         else if (xml.name() == QLatin1String("properties"))
             mMap->mergeProperties(readProperties());
         else if (xml.name() == QLatin1String("tileset"))
@@ -328,7 +335,37 @@ Map *MapReaderPrivate::readMap()
         }
     }
 
-    return mMap.release();
+    return std::move(mMap);
+}
+
+void MapReaderPrivate::readMapEditorSettings(Map &map)
+{
+    Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("editorsettings"));
+
+    while (xml.readNextStartElement()) {
+        if (xml.name() == QLatin1String("chunksize")) {
+            const QXmlStreamAttributes atts = xml.attributes();
+
+            int chunkWidth = atts.value(QLatin1String("width")).toInt();
+            int chunkHeight = atts.value(QLatin1String("height")).toInt();
+
+            chunkWidth = chunkWidth == 0 ? CHUNK_SIZE : qMax(CHUNK_SIZE_MIN, chunkWidth);
+            chunkHeight = chunkHeight == 0 ? CHUNK_SIZE : qMax(CHUNK_SIZE_MIN, chunkHeight);
+
+            map.setChunkSize(QSize(chunkWidth, chunkHeight));
+
+            xml.skipCurrentElement();
+        } else if (xml.name() == QLatin1String("export")) {
+            const QXmlStreamAttributes atts = xml.attributes();
+
+            map.exportFileName = QDir::cleanPath(mPath.filePath(atts.value(QLatin1String("target")).toString()));
+            map.exportFormat = atts.value(QLatin1String("format")).toString();
+
+            xml.skipCurrentElement();
+        } else {
+            readUnknownElement();
+        }
+    }
 }
 
 SharedTileset MapReaderPrivate::readTileset()
@@ -365,7 +402,9 @@ SharedTileset MapReaderPrivate::readTileset()
                 tileset->setBackgroundColor(QColor(bgColorString.toString()));
 
             while (xml.readNextStartElement()) {
-                if (xml.name() == QLatin1String("tile")) {
+                if (xml.name() == QLatin1String("editorsettings")) {
+                    readTilesetEditorSettings(*tileset);
+                } else if (xml.name() == QLatin1String("tile")) {
                     readTilesetTile(*tileset);
                 } else if (xml.name() == QLatin1String("tileoffset")) {
                     const QXmlStreamAttributes oa = xml.attributes();
@@ -416,6 +455,24 @@ SharedTileset MapReaderPrivate::readTileset()
     return tileset;
 }
 
+void MapReaderPrivate::readTilesetEditorSettings(Tileset &tileset)
+{
+    Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("editorsettings"));
+
+    while (xml.readNextStartElement()) {
+        if (xml.name() == QLatin1String("export")) {
+            const QXmlStreamAttributes atts = xml.attributes();
+
+            tileset.exportFileName = QDir::cleanPath(mPath.filePath(atts.value(QLatin1String("target")).toString()));
+            tileset.exportFormat = atts.value(QLatin1String("format")).toString();
+
+            xml.skipCurrentElement();
+        } else {
+            readUnknownElement();
+        }
+    }
+}
+
 void MapReaderPrivate::readTilesetTile(Tileset &tileset)
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("tile"));
@@ -463,7 +520,7 @@ void MapReaderPrivate::readTilesetTile(Tileset &tileset)
                 tileset.setTileImage(tile, image, imageReference.source);
             }
         } else if (xml.name() == QLatin1String("objectgroup")) {
-            ObjectGroup *objectGroup = readObjectGroup();
+            std::unique_ptr<ObjectGroup> objectGroup = readObjectGroup();
             if (objectGroup) {
                 // Migrate properties from the object group to the tile. Since
                 // Tiled 1.1, it is no longer possible to edit the properties
@@ -476,7 +533,7 @@ void MapReaderPrivate::readTilesetTile(Tileset &tileset)
                     objectGroup->setProperties(Properties());
                 }
 
-                tile->setObjectGroup(objectGroup);
+                tile->setObjectGroup(std::move(objectGroup));
             }
         } else if (xml.name() == QLatin1String("animation")) {
             tile->setFrames(readAnimationFrames());
@@ -574,11 +631,11 @@ ImageReference MapReaderPrivate::readImage()
     return image;
 }
 
-ObjectTemplate *MapReaderPrivate::readObjectTemplate()
+std::unique_ptr<ObjectTemplate> MapReaderPrivate::readObjectTemplate()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("template"));
 
-    ObjectTemplate *objectTemplate = new ObjectTemplate;
+    auto objectTemplate = std::make_unique<ObjectTemplate>();
 
     while (xml.readNextStartElement()) {
         if (xml.name() == QLatin1String("object"))
@@ -593,7 +650,7 @@ ObjectTemplate *MapReaderPrivate::readObjectTemplate()
 }
 
 
-Layer *MapReaderPrivate::tryReadLayer()
+std::unique_ptr<Layer> MapReaderPrivate::tryReadLayer()
 {
     Q_ASSERT(xml.isStartElement());
 
@@ -643,9 +700,7 @@ void MapReaderPrivate::readTilesetWangSets(Tileset &tileset)
             QString name = atts.value(QLatin1String("name")).toString();
             int tile = atts.value(QLatin1String("tile")).toInt();
 
-            WangSet *wangSet = new WangSet(&tileset, name, tile);
-
-            tileset.addWangSet(wangSet);
+            auto wangSet = std::make_unique<WangSet>(&tileset, name, tile);
 
             while (xml.readNextStartElement()) {
                 if (xml.name() == QLatin1String("properties")) {
@@ -696,6 +751,8 @@ void MapReaderPrivate::readTilesetWangSets(Tileset &tileset)
                     readUnknownElement();
                 }
             }
+
+            tileset.addWangSet(std::move(wangSet));
         } else {
             readUnknownElement();
         }
@@ -733,7 +790,7 @@ static void readLayerAttributes(Layer &layer,
     layer.setOffset(offset);
 }
 
-TileLayer *MapReaderPrivate::readTileLayer()
+std::unique_ptr<TileLayer> MapReaderPrivate::readTileLayer()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("layer"));
 
@@ -744,7 +801,7 @@ TileLayer *MapReaderPrivate::readTileLayer()
     const int width = atts.value(QLatin1String("width")).toInt();
     const int height = atts.value(QLatin1String("height")).toInt();
 
-    TileLayer *tileLayer = new TileLayer(name, x, y, width, height);
+    auto tileLayer = std::make_unique<TileLayer>(name, x, y, width, height);
     readLayerAttributes(*tileLayer, atts);
 
     while (xml.readNextStartElement()) {
@@ -779,6 +836,8 @@ void MapReaderPrivate::readTileLayerData(TileLayer &tileLayer)
             layerDataFormat = Map::Base64Gzip;
         } else if (compression == QLatin1String("zlib")) {
             layerDataFormat = Map::Base64Zlib;
+        } else if (compression == QLatin1String("zstd")) {
+            layerDataFormat = Map::Base64Zstandard;
         } else {
             xml.raiseError(tr("Compression method '%1' not supported")
                            .arg(compression.toString()));
@@ -928,7 +987,7 @@ Cell MapReaderPrivate::cellForGid(unsigned gid)
     return result;
 }
 
-ObjectGroup *MapReaderPrivate::readObjectGroup()
+std::unique_ptr<ObjectGroup> MapReaderPrivate::readObjectGroup()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("objectgroup"));
 
@@ -937,7 +996,7 @@ ObjectGroup *MapReaderPrivate::readObjectGroup()
     const int x = atts.value(QLatin1String("x")).toInt();
     const int y = atts.value(QLatin1String("y")).toInt();
 
-    ObjectGroup *objectGroup = new ObjectGroup(name, x, y);
+    auto objectGroup = std::make_unique<ObjectGroup>(name, x, y);
     readLayerAttributes(*objectGroup, atts);
 
     const QString color = atts.value(QLatin1String("color")).toString();
@@ -948,7 +1007,6 @@ ObjectGroup *MapReaderPrivate::readObjectGroup()
         QString value = atts.value(QLatin1String("draworder")).toString();
         ObjectGroup::DrawOrder drawOrder = drawOrderFromString(value);
         if (drawOrder == ObjectGroup::UnknownOrder) {
-            delete objectGroup;
             xml.raiseError(tr("Invalid draw order: %1").arg(value));
             return nullptr;
         }
@@ -967,7 +1025,7 @@ ObjectGroup *MapReaderPrivate::readObjectGroup()
     return objectGroup;
 }
 
-ImageLayer *MapReaderPrivate::readImageLayer()
+std::unique_ptr<ImageLayer> MapReaderPrivate::readImageLayer()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("imagelayer"));
 
@@ -976,7 +1034,7 @@ ImageLayer *MapReaderPrivate::readImageLayer()
     const int x = atts.value(QLatin1String("x")).toInt();
     const int y = atts.value(QLatin1String("y")).toInt();
 
-    ImageLayer *imageLayer = new ImageLayer(name, x, y);
+    auto imageLayer = std::make_unique<ImageLayer>(name, x, y);
     readLayerAttributes(*imageLayer, atts);
 
     // Image layer pixel position moved from x/y to offsetx/offsety for
@@ -1017,7 +1075,7 @@ void MapReaderPrivate::readImageLayerImage(ImageLayer &imageLayer)
     xml.skipCurrentElement();
 }
 
-MapObject *MapReaderPrivate::readObject()
+std::unique_ptr<MapObject> MapReaderPrivate::readObject()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("object"));
 
@@ -1036,7 +1094,7 @@ MapObject *MapReaderPrivate::readObject()
     const QPointF pos(x, y);
     const QSizeF size(width, height);
 
-    MapObject *object = new MapObject(name, type, pos, size);
+    auto object = std::make_unique<MapObject>(name, type, pos, size);
 
     if (!templateFileName.isEmpty()) { // This object is a template instance
         const QString absoluteFileName = p->resolveReference(templateFileName, mPath);
@@ -1121,10 +1179,10 @@ QPolygonF MapReaderPrivate::readPolygon()
             break;
         }
 
-        const qreal x = point.left(commaPos).toDouble(&ok);
+        const qreal x = point.leftRef(commaPos).toDouble(&ok);
         if (!ok)
             break;
-        const qreal y = point.mid(commaPos + 1).toDouble(&ok);
+        const qreal y = point.midRef(commaPos + 1).toDouble(&ok);
         if (!ok)
             break;
 
@@ -1177,6 +1235,8 @@ TextData MapReaderPrivate::readObjectText()
         alignment |= Qt::AlignHCenter;
     else if (hAlignString == QLatin1String("right"))
         alignment |= Qt::AlignRight;
+    else if (hAlignString == QLatin1String("justify"))
+        alignment |= Qt::AlignJustify;
     else
         alignment |= Qt::AlignLeft;
 
@@ -1195,7 +1255,7 @@ TextData MapReaderPrivate::readObjectText()
     return textData;
 }
 
-GroupLayer *MapReaderPrivate::readGroupLayer()
+std::unique_ptr<GroupLayer> MapReaderPrivate::readGroupLayer()
 {
     Q_ASSERT(xml.isStartElement() && xml.name() == QLatin1String("group"));
 
@@ -1204,12 +1264,12 @@ GroupLayer *MapReaderPrivate::readGroupLayer()
     const int x = atts.value(QLatin1String("x")).toInt();
     const int y = atts.value(QLatin1String("y")).toInt();
 
-    GroupLayer *groupLayer = new GroupLayer(name, x, y);
+    auto groupLayer = std::make_unique<GroupLayer>(name, x, y);
     readLayerAttributes(*groupLayer, atts);
 
     while (xml.readNextStartElement()) {
-        if (Layer *layer = tryReadLayer())
-            groupLayer->addLayer(layer);
+        if (std::unique_ptr<Layer> layer = tryReadLayer())
+            groupLayer->addLayer(std::move(layer));
         else if (xml.name() == QLatin1String("properties"))
             groupLayer->mergeProperties(readProperties());
         else
@@ -1300,12 +1360,12 @@ MapReader::~MapReader()
     delete d;
 }
 
-Map *MapReader::readMap(QIODevice *device, const QString &path)
+std::unique_ptr<Map> MapReader::readMap(QIODevice *device, const QString &path)
 {
     return d->readMap(device, path);
 }
 
-Map *MapReader::readMap(const QString &fileName)
+std::unique_ptr<Map> MapReader::readMap(const QString &fileName)
 {
     QFile file(fileName);
     if (!d->openFile(&file))
@@ -1329,26 +1389,21 @@ SharedTileset MapReader::readTileset(const QString &fileName)
     if (!d->openFile(&file))
         return SharedTileset();
 
-    SharedTileset tileset = readTileset(&file, QFileInfo(fileName).absolutePath());
-    if (tileset)
-        tileset->setFileName(fileName);
-
-    return tileset;
+    return readTileset(&file, QFileInfo(fileName).absolutePath());
 }
 
-ObjectTemplate *MapReader::readObjectTemplate(QIODevice *device, const QString &path)
+std::unique_ptr<ObjectTemplate> MapReader::readObjectTemplate(QIODevice *device, const QString &path)
 {
     return d->readObjectTemplate(device, path);
 }
 
-ObjectTemplate *MapReader::readObjectTemplate(const QString &fileName)
+std::unique_ptr<ObjectTemplate> MapReader::readObjectTemplate(const QString &fileName)
 {
     QFile file(fileName);
     if (!d->openFile(&file))
         return nullptr;
 
-    ObjectTemplate *objectTemplate = readObjectTemplate(&file, QFileInfo(fileName).absolutePath());
-
+    auto objectTemplate = readObjectTemplate(&file, QFileInfo(fileName).absolutePath());
     if (objectTemplate)
         objectTemplate->setFileName(fileName);
 
