@@ -47,8 +47,9 @@
 
 #include <QtMath>
 
+#include "qtcompat_p.h"
+
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 static bool isTileObject(MapObject *mapObject)
 {
@@ -84,15 +85,15 @@ AbstractObjectTool::AbstractObjectTool(const QString &name,
     : AbstractTool(name, icon, shortcut, parent)
     , mMapScene(nullptr)
 {
-    QIcon flipHorizontalIcon(QLatin1String(":images/24x24/flip-horizontal.png"));
-    QIcon flipVerticalIcon(QLatin1String(":images/24x24/flip-vertical.png"));
-    QIcon rotateLeftIcon(QLatin1String(":images/24x24/rotate-left.png"));
-    QIcon rotateRightIcon(QLatin1String(":images/24x24/rotate-right.png"));
+    QIcon flipHorizontalIcon(QLatin1String(":images/24/flip-horizontal.png"));
+    QIcon flipVerticalIcon(QLatin1String(":images/24/flip-vertical.png"));
+    QIcon rotateLeftIcon(QLatin1String(":images/24/rotate-left.png"));
+    QIcon rotateRightIcon(QLatin1String(":images/24/rotate-right.png"));
 
-    flipHorizontalIcon.addFile(QLatin1String(":images/32x32/flip-horizontal.png"));
-    flipVerticalIcon.addFile(QLatin1String(":images/32x32/flip-vertical.png"));
-    rotateLeftIcon.addFile(QLatin1String(":images/32x32/rotate-left.png"));
-    rotateRightIcon.addFile(QLatin1String(":images/32x32/rotate-right.png"));
+    flipHorizontalIcon.addFile(QLatin1String(":images/32/flip-horizontal.png"));
+    flipVerticalIcon.addFile(QLatin1String(":images/32/flip-vertical.png"));
+    rotateLeftIcon.addFile(QLatin1String(":images/32/rotate-left.png"));
+    rotateRightIcon.addFile(QLatin1String(":images/32/rotate-right.png"));
 
     mFlipHorizontal = new QAction(this);
     mFlipHorizontal->setIcon(flipHorizontalIcon);
@@ -166,7 +167,7 @@ void AbstractObjectTool::mouseMoved(const QPointF &pos,
 void AbstractObjectTool::mousePressed(QGraphicsSceneMouseEvent *event)
 {
     if (event->button() == Qt::RightButton) {
-        showContextMenu(topMostObjectItemAt(event->scenePos()),
+        showContextMenu(topMostMapObjectAt(event->scenePos()),
                         event->screenPos());
     }
 }
@@ -175,13 +176,13 @@ void AbstractObjectTool::languageChanged()
 {
     mFlipHorizontal->setToolTip(tr("Flip Horizontally"));
     mFlipVertical->setToolTip(tr("Flip Vertically"));
-    mRotateLeft->setToolTip(QCoreApplication::translate("Tiled::Internal::StampActions", "Rotate Left"));
-    mRotateRight->setToolTip(QCoreApplication::translate("Tiled::Internal::StampActions", "Rotate Right"));
+    mRotateLeft->setToolTip(QCoreApplication::translate("Tiled::StampActions", "Rotate Left"));
+    mRotateRight->setToolTip(QCoreApplication::translate("Tiled::StampActions", "Rotate Right"));
 
-    mFlipHorizontal->setShortcut(QKeySequence(QCoreApplication::translate("Tiled::Internal::StampActions", "X")));
-    mFlipVertical->setShortcut(QKeySequence(QCoreApplication::translate("Tiled::Internal::StampActions", "Y")));
-    mRotateLeft->setShortcut(QKeySequence(QCoreApplication::translate("Tiled::Internal::StampActions", "Shift+Z")));
-    mRotateRight->setShortcut(QKeySequence(QCoreApplication::translate("Tiled::Internal::StampActions", "Z")));
+    mFlipHorizontal->setShortcut(Qt::Key_X);
+    mFlipVertical->setShortcut(Qt::Key_Y);
+    mRotateLeft->setShortcut(Qt::SHIFT + Qt::Key_Z);
+    mRotateRight->setShortcut(Qt::Key_Z);
 }
 
 void AbstractObjectTool::populateToolBar(QToolBar *toolBar)
@@ -205,27 +206,33 @@ ObjectGroup *AbstractObjectTool::currentObjectGroup() const
     return dynamic_cast<ObjectGroup*>(mapDocument()->currentLayer());
 }
 
-QList<MapObjectItem*> AbstractObjectTool::objectItemsAt(const QPointF &pos) const
+QList<MapObject*> AbstractObjectTool::mapObjectsAt(const QPointF &pos) const
 {
     const QList<QGraphicsItem *> &items = mMapScene->items(pos);
 
-    QList<MapObjectItem*> objectList;
+    QList<MapObject*> objectList;
     for (auto item : items) {
+        if (!item->isEnabled())
+            continue;
+
         MapObjectItem *objectItem = qgraphicsitem_cast<MapObjectItem*>(item);
         if (objectItem && objectItem->mapObject()->objectGroup()->isUnlocked())
-            objectList.append(objectItem);
+            objectList.append(objectItem->mapObject());
     }
     return objectList;
 }
 
-MapObjectItem *AbstractObjectTool::topMostObjectItemAt(const QPointF &pos) const
+MapObject *AbstractObjectTool::topMostMapObjectAt(const QPointF &pos) const
 {
     const QList<QGraphicsItem *> &items = mMapScene->items(pos);
 
     for (QGraphicsItem *item : items) {
+        if (!item->isEnabled())
+            continue;
+
         MapObjectItem *objectItem = qgraphicsitem_cast<MapObjectItem*>(item);
         if (objectItem && objectItem->mapObject()->objectGroup()->isUnlocked())
-            return objectItem;
+            return objectItem->mapObject();
     }
     return nullptr;
 }
@@ -257,7 +264,7 @@ void AbstractObjectTool::resetTileSize()
     if (!commands.isEmpty()) {
         QUndoStack *undoStack = mapDocument()->undoStack();
         undoStack->beginMacro(tr("Reset Tile Size"));
-        for (auto command : commands)
+        for (auto command : qAsConst(commands))
             undoStack->push(command);
         undoStack->endMacro();
     }
@@ -275,12 +282,12 @@ static QString saveObjectTemplate(const MapObject *mapObject)
     if (!mapObject->name().isEmpty())
         suggestedFileName += mapObject->name();
     else
-        suggestedFileName += QCoreApplication::translate("Tiled::Internal::MainWindow", "untitled");
+        suggestedFileName += QCoreApplication::translate("Tiled::MainWindow", "untitled");
     suggestedFileName += QLatin1String(".tx");
 
     QWidget *parent = DocumentManager::instance()->widget()->window();
     QString fileName = QFileDialog::getSaveFileName(parent,
-                                                    QCoreApplication::translate("Tiled::Internal::MainWindow", "Save Template"),
+                                                    QCoreApplication::translate("Tiled::MainWindow", "Save Template"),
                                                     suggestedFileName,
                                                     filter,
                                                     &selectedFilter);
@@ -294,7 +301,7 @@ static QString saveObjectTemplate(const MapObject *mapObject)
     objectTemplate.setObject(mapObject);
 
     if (!format->write(&objectTemplate, fileName)) {
-        QMessageBox::critical(nullptr, QCoreApplication::translate("Tiled::Internal::MainWindow", "Error Saving Template"),
+        QMessageBox::critical(nullptr, QCoreApplication::translate("Tiled::MainWindow", "Error Saving Template"),
                               format->errorString());
         return QString();
     }
@@ -340,7 +347,7 @@ void AbstractObjectTool::detachSelectedObjects()
     auto changeMapObjectCommand = new DetachObjects(currentMapDocument, templateInstances);
 
     // Add any missing tileset used by the templates to the map map before detaching
-    for (SharedTileset sharedTileset : sharedTilesets) {
+    for (const SharedTileset &sharedTileset : qAsConst(sharedTilesets)) {
         if (!currentMapDocument->map()->tilesets().contains(sharedTileset))
             new AddTileset(currentMapDocument, sharedTileset, changeMapObjectCommand);
     }
@@ -431,51 +438,51 @@ void AbstractObjectTool::lowerToBottom()
  * Shows the context menu for map objects. The menu allows you to duplicate and
  * remove the map objects, or to edit their properties.
  */
-void AbstractObjectTool::showContextMenu(MapObjectItem *clickedObjectItem,
+void AbstractObjectTool::showContextMenu(MapObject *clickedObject,
                                          QPoint screenPos)
 {
     const QList<MapObject*> &selectedObjects = mapDocument()->selectedObjects();
 
-    if (clickedObjectItem && !selectedObjects.contains(clickedObjectItem->mapObject()))
-        mapDocument()->setSelectedObjects({ clickedObjectItem->mapObject() });
+    if (clickedObject && !selectedObjects.contains(clickedObject))
+        mapDocument()->setSelectedObjects({ clickedObject });
 
     if (selectedObjects.isEmpty())
         return;
 
     QMenu menu;
     QAction *duplicateAction = menu.addAction(tr("Duplicate %n Object(s)", "", selectedObjects.size()),
-                                              this, SLOT(duplicateObjects()));
+                                              this, &AbstractObjectTool::duplicateObjects);
     QAction *removeAction = menu.addAction(tr("Remove %n Object(s)", "", selectedObjects.size()),
-                                           this, SLOT(removeObjects()));
+                                           this, &AbstractObjectTool::removeObjects);
 
-    duplicateAction->setIcon(QIcon(QLatin1String(":/images/16x16/stock-duplicate-16.png")));
-    removeAction->setIcon(QIcon(QLatin1String(":/images/16x16/edit-delete.png")));
+    duplicateAction->setIcon(QIcon(QLatin1String(":/images/16/stock-duplicate-16.png")));
+    removeAction->setIcon(QIcon(QLatin1String(":/images/16/edit-delete.png")));
 
     bool anyTileObjectSelected = std::any_of(selectedObjects.begin(),
                                              selectedObjects.end(),
                                              isTileObject);
 
     if (anyTileObjectSelected) {
-        auto resetTileSizeAction = menu.addAction(tr("Reset Tile Size"), this, SLOT(resetTileSize()));
+        auto resetTileSizeAction = menu.addAction(tr("Reset Tile Size"), this, &AbstractObjectTool::resetTileSize);
         resetTileSizeAction->setEnabled(std::any_of(selectedObjects.begin(),
                                                     selectedObjects.end(),
                                                     isResizedTileObject));
 
-        auto changeTileAction = menu.addAction(tr("Replace Tile"), this, SLOT(changeTile()));
+        auto changeTileAction = menu.addAction(tr("Replace Tile"), this, &AbstractObjectTool::changeTile);
         changeTileAction->setEnabled(tile() && (!selectedObjects.first()->isTemplateBase() ||
                                                 tile()->tileset()->isExternal()));
     }
 
     // Create action for replacing an object with a template
+    auto replaceTemplateAction = menu.addAction(tr("Replace With Template"), this, &AbstractObjectTool::replaceObjectsWithTemplate);
     auto selectedTemplate = objectTemplate();
-    auto replaceTemplateAction = menu.addAction(tr("Replace With Template"), this, SLOT(replaceObjectsWithTemplate()));
 
     if (selectedTemplate) {
         QString name = QFileInfo(selectedTemplate->fileName()).fileName();
         replaceTemplateAction->setText(tr("Replace With Template \"%1\"").arg(name));
-    } else {
-        replaceTemplateAction->setEnabled(false);
     }
+    if (!selectedTemplate || !mapDocument()->templateAllowed(selectedTemplate))
+        replaceTemplateAction->setEnabled(false);
 
     if (selectedObjects.size() == 1) {
         MapObject *currentObject = selectedObjects.first();
@@ -484,7 +491,7 @@ void AbstractObjectTool::showContextMenu(MapObjectItem *clickedObjectItem,
             const Cell cell = selectedObjects.first()->cell();
             // Saving objects with embedded tilesets is disabled
             if (cell.isEmpty() || cell.tileset()->isExternal())
-                menu.addAction(tr("Save As Template"), this, SLOT(saveSelectedObject()));
+                menu.addAction(tr("Save As Template"), this, &AbstractObjectTool::saveSelectedObject);
         }
 
         if (currentObject->isTemplateBase()) { // Hide this operations for template base
@@ -499,40 +506,42 @@ void AbstractObjectTool::showContextMenu(MapObjectItem *clickedObjectItem,
                                                    isTemplateInstance);
 
     if (anyTemplateInstanceSelected) {
-        menu.addAction(tr("Detach"), this, SLOT(detachSelectedObjects()));
+        menu.addAction(tr("Detach"), this, &AbstractObjectTool::detachSelectedObjects);
 
-        auto resetToTemplateAction = menu.addAction(tr("Reset Template Instance(s)"), this, SLOT(resetInstances()));
+        auto resetToTemplateAction = menu.addAction(tr("Reset Template Instance(s)"), this, &AbstractObjectTool::resetInstances);
         resetToTemplateAction->setEnabled(std::any_of(selectedObjects.begin(),
                                                       selectedObjects.end(),
                                                       isChangedTemplateInstance));
     }
 
     menu.addSeparator();
-    menu.addAction(tr("Flip Horizontally"), this, SLOT(flipHorizontally()), QKeySequence(tr("X")));
-    menu.addAction(tr("Flip Vertically"), this, SLOT(flipVertically()), QKeySequence(tr("Y")));
+    menu.addAction(tr("Flip Horizontally"), this, &AbstractObjectTool::flipHorizontally, Qt::Key_X);
+    menu.addAction(tr("Flip Vertically"), this, &AbstractObjectTool::flipVertically, Qt::Key_Y);
 
-    ObjectGroup *objectGroup = RaiseLowerHelper::sameObjectGroup(selectedObjects);
-    if (objectGroup && objectGroup->drawOrder() == ObjectGroup::IndexOrder) {
+    ObjectGroup *sameObjectGroup = RaiseLowerHelper::sameObjectGroup(selectedObjects);
+    if (sameObjectGroup && sameObjectGroup->drawOrder() == ObjectGroup::IndexOrder) {
         menu.addSeparator();
-        menu.addAction(tr("Raise Object"), this, SLOT(raise()), QKeySequence(tr("PgUp")));
-        menu.addAction(tr("Lower Object"), this, SLOT(lower()), QKeySequence(tr("PgDown")));
-        menu.addAction(tr("Raise Object to Top"), this, SLOT(raiseToTop()), QKeySequence(tr("Home")));
-        menu.addAction(tr("Lower Object to Bottom"), this, SLOT(lowerToBottom()), QKeySequence(tr("End")));
+        menu.addAction(tr("Raise Object"), this, &AbstractObjectTool::raise, Qt::Key_PageUp);
+        menu.addAction(tr("Lower Object"), this, &AbstractObjectTool::lower, Qt::Key_PageDown);
+        menu.addAction(tr("Raise Object to Top"), this, &AbstractObjectTool::raiseToTop, Qt::Key_Home);
+        menu.addAction(tr("Lower Object to Bottom"), this, &AbstractObjectTool::lowerToBottom, Qt::Key_End);
     }
 
-    const QList<ObjectGroup*> objectGroups = mapDocument()->map()->objectGroups();
-    if (objectGroups.size() > 1) {
+    auto objectGroups = mapDocument()->map()->objectGroups();
+    if (!objectGroups.isEmpty()) {
         menu.addSeparator();
         QMenu *moveToLayerMenu = menu.addMenu(tr("Move %n Object(s) to Layer",
                                                  "", selectedObjects.size()));
-        for (ObjectGroup *objectGroup : objectGroups) {
+        for (Layer *layer : objectGroups) {
+            ObjectGroup *objectGroup = static_cast<ObjectGroup*>(layer);
             QAction *action = moveToLayerMenu->addAction(objectGroup->name());
             action->setData(QVariant::fromValue(objectGroup));
+            action->setEnabled(objectGroup != sameObjectGroup);
         }
     }
 
     menu.addSeparator();
-    QIcon propIcon(QLatin1String(":images/16x16/document-properties.png"));
+    QIcon propIcon(QLatin1String(":images/16/document-properties.png"));
     QAction *propertiesAction = menu.addAction(propIcon,
                                                tr("Object &Properties..."));
 
@@ -550,6 +559,9 @@ void AbstractObjectTool::showContextMenu(MapObjectItem *clickedObjectItem,
         return;
     }
 
-    if (ObjectGroup *objectGroup = action->data().value<ObjectGroup*>())
+    if (ObjectGroup *objectGroup = action->data().value<ObjectGroup*>()) {
+        const auto selectedObjectsCopy = selectedObjects;
         mapDocument()->moveObjectsToGroup(selectedObjects, objectGroup);
+        mapDocument()->setSelectedObjects(selectedObjectsCopy);
+    }
 }
