@@ -20,8 +20,9 @@
 
 #include "abstracttileselectiontool.h"
 
-#include "mapdocument.h"
+#include "brushitem.h"
 #include "changeselectedarea.h"
+#include "mapdocument.h"
 
 #include <QAction>
 #include <QActionGroup>
@@ -29,7 +30,6 @@
 #include <QToolBar>
 
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 AbstractTileSelectionTool::AbstractTileSelectionTool(const QString &name,
                                                      const QIcon &icon,
@@ -39,10 +39,10 @@ AbstractTileSelectionTool::AbstractTileSelectionTool(const QString &name,
     , mSelectionMode(Replace)
     , mDefaultMode(Replace)
 {
-    QIcon replaceIcon(QLatin1String(":images/16x16/selection-replace.png"));
-    QIcon addIcon(QLatin1String(":images/16x16/selection-add.png"));
-    QIcon subtractIcon(QLatin1String(":images/16x16/selection-subtract.png"));
-    QIcon intersectIcon(QLatin1String(":images/16x16/selection-intersect.png"));
+    QIcon replaceIcon(QLatin1String(":images/16/selection-replace.png"));
+    QIcon addIcon(QLatin1String(":images/16/selection-add.png"));
+    QIcon subtractIcon(QLatin1String(":images/16/selection-subtract.png"));
+    QIcon intersectIcon(QLatin1String(":images/16/selection-intersect.png"));
 
     mReplace = new QAction(this);
     mReplace->setIcon(replaceIcon);
@@ -68,13 +68,13 @@ AbstractTileSelectionTool::AbstractTileSelectionTool(const QString &name,
     mActionGroup->addAction(mIntersect);
 
     connect(mReplace, &QAction::triggered,
-            [this]() { mSelectionMode = mDefaultMode = Replace; });
+            [this] { mSelectionMode = mDefaultMode = Replace; });
     connect(mAdd, &QAction::triggered,
-            [this]() { mSelectionMode = mDefaultMode = Add; });
+            [this] { mSelectionMode = mDefaultMode = Add; });
     connect(mSubtract, &QAction::triggered,
-            [this]() { mSelectionMode = mDefaultMode = Subtract; });
+            [this] { mSelectionMode = mDefaultMode = Subtract; });
     connect(mIntersect, &QAction::triggered,
-            [this]() { mSelectionMode = mDefaultMode = Intersect; });
+            [this] { mSelectionMode = mDefaultMode = Intersect; });
 
     AbstractTileSelectionTool::languageChanged();
 }
@@ -83,29 +83,31 @@ void AbstractTileSelectionTool::mousePressed(QGraphicsSceneMouseEvent *event)
 {
     const Qt::MouseButton button = event->button();
 
-    if (button != Qt::LeftButton && button != Qt::RightButton)
-        return;
+    if (button == Qt::LeftButton || (button == Qt::RightButton && event->modifiers() == Qt::NoModifier)) {
+        MapDocument *document = mapDocument();
+        QRegion selection;
 
-    MapDocument *document = mapDocument();
+        // Left button modifies selection, right button clears selection
+        if (button == Qt::LeftButton) {
+            selection = document->selectedArea();
 
-    QRegion selection;
-
-    // Left button modifies selection, right button clears selection
-    if (button == Qt::LeftButton) {
-        selection = document->selectedArea();
-
-        switch (mSelectionMode) {
-        case Replace:   selection = mSelectedRegion; break;
-        case Add:       selection += mSelectedRegion; break;
-        case Subtract:  selection -= mSelectedRegion; break;
-        case Intersect: selection &= mSelectedRegion; break;
+            switch (mSelectionMode) {
+            case Replace:   selection = mSelectedRegion; break;
+            case Add:       selection += mSelectedRegion; break;
+            case Subtract:  selection -= mSelectedRegion; break;
+            case Intersect: selection &= mSelectedRegion; break;
+            }
         }
+
+        if (selection != document->selectedArea()) {
+            QUndoCommand *cmd = new ChangeSelectedArea(document, selection);
+            document->undoStack()->push(cmd);
+        }
+
+        return;
     }
 
-    if (selection != document->selectedArea()) {
-        QUndoCommand *cmd = new ChangeSelectedArea(document, selection);
-        document->undoStack()->push(cmd);
-    }
+    AbstractTileTool::mousePressed(event);
 }
 
 void AbstractTileSelectionTool::mouseReleased(QGraphicsSceneMouseEvent *)
@@ -146,3 +148,9 @@ void AbstractTileSelectionTool::populateToolBar(QToolBar *toolBar)
     toolBar->addAction(mSubtract);
     toolBar->addAction(mIntersect);
 }
+
+// Override to ignore whether the current layer is a visible tile layer
+void AbstractTileSelectionTool::updateBrushVisibility()
+{
+    brushItem()->setVisible(isBrushVisible());
+}
