@@ -27,7 +27,6 @@ class QAction;
 class QActionGroup;
 
 namespace Tiled {
-namespace Internal {
 
 class MapDocument;
 
@@ -58,11 +57,13 @@ protected:
         Intersect
     };
 
-    SelectionMode selectionMode() { return mSelectionMode; }
+    SelectionMode selectionMode() const { return mSelectionMode; }
 
-    QRegion selectedRegion() { return mSelectedRegion; }
+    QRegion selectedRegion() const { return mSelectedRegion; }
     void setSelectedRegion(QRegion region) { mSelectedRegion = region; }
 
+    void updateBrushVisibility() override;
+
 private:
     SelectionMode mSelectionMode;
     SelectionMode mDefaultMode;
@@ -76,5 +77,4 @@ private:
     QActionGroup *mActionGroup;
 };
 
-} // namespace Internal
 } // namespace Tiled
