@@ -21,6 +21,8 @@
 
 #pragma once
 
+#include "changeevents.h"
+
 #include <QCursor>
 #include <QGraphicsSceneMouseEvent>
 #include <QIcon>
@@ -39,10 +41,9 @@ class Layer;
 class Tile;
 class ObjectTemplate;
 
-namespace Internal {
-
 class MapDocument;
 class MapScene;
+class ToolManager;
 
 /**
  * An abstraction of any kind of tool used to edit the map.
@@ -57,9 +58,9 @@ class AbstractTool : public QObject
 {
     Q_OBJECT
 
-    Q_PROPERTY(QString name READ name WRITE setName)
-    Q_PROPERTY(QIcon icon READ icon WRITE setIcon)
-    Q_PROPERTY(QKeySequence shortcut READ shortcut WRITE setShortcut)
+    Q_PROPERTY(QString name READ name WRITE setName NOTIFY changed)
+    Q_PROPERTY(QIcon icon READ icon WRITE setIcon NOTIFY changed)
+    Q_PROPERTY(QKeySequence shortcut READ shortcut WRITE setShortcut NOTIFY changed)
     Q_PROPERTY(QString statusInfo READ statusInfo WRITE setStatusInfo NOTIFY statusInfoChanged)
     Q_PROPERTY(QCursor cursor READ cursor WRITE setCursor NOTIFY cursorChanged)
     Q_PROPERTY(bool enabled READ isEnabled WRITE setEnabled NOTIFY enabledChanged)
@@ -91,8 +92,8 @@ public:
     bool isEnabled() const;
     void setEnabled(bool enabled);
 
+    ToolManager *toolManager() const;
     Tile *tile() const;
-
     ObjectTemplate *objectTemplate() const;
 
     /**
@@ -158,12 +159,11 @@ public:
 
     virtual void populateToolBar(QToolBar*) {}
 
-public slots:
     void setMapDocument(MapDocument *mapDocument);
-    void setTile(Tile *tile);
-    void setObjectTemplate(ObjectTemplate *objectTemplate);
 
 protected:
+    virtual void changeEvent(const ChangeEvent &event);
+
     /**
      * Can be used to respond to the map document changing.
      */
@@ -178,7 +178,6 @@ protected:
 
     Layer *currentLayer() const;
 
-protected slots:
     /**
      * By default, this function is called after the current map has changed
      * and when the current layer changes. It can be overridden to implement
@@ -189,20 +188,22 @@ protected slots:
     virtual void updateEnabledState();
 
 signals:
+    void changed();
     void statusInfoChanged(const QString &statusInfo);
     void cursorChanged(const QCursor &cursor);
     void enabledChanged(bool enabled);
 
 private:
+    friend class ToolManager;
+
     QString mName;
     QIcon mIcon;
     QKeySequence mShortcut;
     QString mStatusInfo;
     QCursor mCursor;
     bool mEnabled;
-    Tile *mTile;
-    ObjectTemplate *mObjectTemplate;
 
+    ToolManager *mToolManager;
     MapDocument *mMapDocument;
 };
 
@@ -212,30 +213,16 @@ inline QString AbstractTool::name() const
     return mName;
 }
 
-inline void AbstractTool::setName(const QString &name)
-{
-    mName = name;
-}
-
 inline QIcon AbstractTool::icon() const
 {
     return mIcon;
 }
 
-inline void AbstractTool::setIcon(const QIcon &icon)
-{
-    mIcon = icon;
-}
-
 inline QKeySequence AbstractTool::shortcut() const
 {
     return mShortcut;
 }
 
-inline void AbstractTool::setShortcut(const QKeySequence &shortcut)
-{
-    mShortcut = shortcut;
-}
 
 inline QString AbstractTool::statusInfo() const
 {
@@ -252,27 +239,15 @@ inline bool AbstractTool::isEnabled() const
     return mEnabled;
 }
 
-inline Tile *AbstractTool::tile() const
+/**
+ * Returns the ToolManager with which this tool is registered, if any.
+ */
+inline ToolManager *AbstractTool::toolManager() const
 {
-    return mTile;
+    return mToolManager;
 }
 
-inline void AbstractTool::setTile(Tile *tile)
-{
-    mTile = tile;
-}
-
-inline ObjectTemplate *AbstractTool::objectTemplate() const
-{
-    return mObjectTemplate;
-}
-
-inline void AbstractTool::setObjectTemplate(ObjectTemplate *objectTemplate)
-{
-    mObjectTemplate = objectTemplate;
-}
-
-} // namespace Internal
 } // namespace Tiled
 
-Q_DECLARE_METATYPE(Tiled::Internal::AbstractTool*)
+Q_DECLARE_METATYPE(Tiled::AbstractTool*)
+Q_DECLARE_INTERFACE(Tiled::AbstractTool, "org.mapeditor.AbstractTool")
