@@ -24,22 +24,27 @@
 #include "utils.h"
 
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 CreateEllipseObjectTool::CreateEllipseObjectTool(QObject *parent)
     : CreateScalableObjectTool(parent)
 {
-    QIcon icon(QLatin1String(":images/24x24/insert-ellipse.png"));
-    icon.addFile(QLatin1String(":images/48x48/insert-ellipse.png"));
+    QIcon icon(QLatin1String(":images/24/insert-ellipse.png"));
+    icon.addFile(QLatin1String(":images/48/insert-ellipse.png"));
     setIcon(icon);
+    setShortcut(Qt::Key_C);
     Utils::setThemeIcon(this, "insert-ellipse");
-    languageChanged();
+    languageChangedImpl();
 }
 
 void CreateEllipseObjectTool::languageChanged()
+{
+    CreateScalableObjectTool::languageChanged();
+    languageChangedImpl();
+}
+
+void CreateEllipseObjectTool::languageChangedImpl()
 {
     setName(tr("Insert Ellipse"));
-    setShortcut(QKeySequence(tr("C")));
 }
 
 MapObject *CreateEllipseObjectTool::createNewMapObject()
