@@ -22,12 +22,12 @@
 
 #include "abstractobjecttool.h"
 
+#include <memory>
+
 namespace Tiled {
 
 class Tile;
 
-namespace Internal {
-
 class MapObjectItem;
 class ObjectGroupItem;
 
@@ -44,29 +44,52 @@ public:
 
     void keyPressed(QKeyEvent *event) override;
     void mouseEntered() override;
+    void mouseLeft() override;
     void mouseMoved(const QPointF &pos,
                     Qt::KeyboardModifiers modifiers) override;
     void mousePressed(QGraphicsSceneMouseEvent *event) override;
     void mouseReleased(QGraphicsSceneMouseEvent *event) override;
+    void modifiersChanged(Qt::KeyboardModifiers modifiers) override;
 
 protected:
+    void changeEvent(const ChangeEvent &event) override;
+
+    void updateEnabledState() override;
+
+    enum State {
+        Idle,
+        Preview,
+        CreatingObject,
+    };
+
     virtual void mouseMovedWhileCreatingObject(const QPointF &pos,
                                                Qt::KeyboardModifiers modifiers);
-    virtual void mousePressedWhileCreatingObject(QGraphicsSceneMouseEvent *event);
-    virtual void mouseReleasedWhileCreatingObject(QGraphicsSceneMouseEvent *event);
-
 
     virtual bool startNewMapObject(const QPointF &pos, ObjectGroup *objectGroup);
     virtual MapObject *createNewMapObject() = 0;
     virtual void cancelNewMapObject();
     virtual void finishNewMapObject();
+    virtual std::unique_ptr<MapObject> clearNewMapObjectItem();
 
-    MapObject *clearNewMapObjectItem();
-    ObjectGroup *mNewMapObjectGroup;
-    ObjectGroupItem *mObjectGroupItem;
-    MapObjectItem *mNewMapObjectItem;
-    MapObjectItem *mOverlayPolygonItem;
+    State state() const { return mState; }
+    void setState(State state) { mState = state; }
+
+    ObjectGroup *newMapObjectGroup() { return mNewMapObjectGroup.get(); }
+    ObjectGroupItem *objectGroupItem() { return mObjectGroupItem.get(); }
+
+    MapObjectItem *mNewMapObjectItem;   // owned by mObjectGroupItem if set
+
+private:
+    void objectGroupChanged(ObjectGroup *objectGroup);
+
+    void tryCreatePreview(const QPointF &scenePos,
+                          Qt::KeyboardModifiers modifiers);
+
+    State mState = Idle;
+    QPointF mLastScenePos;
+    Qt::KeyboardModifiers mLastModifiers = Qt::NoModifier;
+    std::unique_ptr<ObjectGroup> mNewMapObjectGroup;
+    std::unique_ptr<ObjectGroupItem> mObjectGroupItem;
 };
 
-} // namespace Internal
 } // namespace Tiled
