@@ -23,7 +23,6 @@
 #include "createobjecttool.h"
 
 namespace Tiled {
-namespace Internal {
 
 class CreateScalableObjectTool : public CreateObjectTool
 {
@@ -37,12 +36,9 @@ protected:
 
     void mouseMovedWhileCreatingObject(const QPointF &pos,
                                        Qt::KeyboardModifiers modifiers) override;
-    void mousePressedWhileCreatingObject(QGraphicsSceneMouseEvent *event) override;
-    void mouseReleasedWhileCreatingObject(QGraphicsSceneMouseEvent *event) override;
 
 private:
     QPointF mStartPos;
 };
 
-} // namespace Internal
 } // namespace Tiled
