@@ -30,53 +30,18 @@
 #include "utils.h"
 
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 CreateTemplateTool::CreateTemplateTool(QObject *parent)
     : CreateObjectTool(parent)
 {
-    QIcon icon(QLatin1String(":images/24x24/insert-template.png"));
-    icon.addFile(QLatin1String(":images/48x48/insert-template.png"));
+    QIcon icon(QLatin1String(":images/24/insert-template.png"));
+    icon.addFile(QLatin1String(":images/48/insert-template.png"));
     setIcon(icon);
+    setShortcut(Qt::Key_V);
     Utils::setThemeIcon(this, "insert-template");
     languageChangedImpl();
 }
 
-void CreateTemplateTool::mouseMovedWhileCreatingObject(const QPointF &pos, Qt::KeyboardModifiers modifiers)
-{
-    const MapRenderer *renderer = mapDocument()->renderer();
-
-    QPointF pixelCoords = renderer->screenToPixelCoords(pos);
-
-    SnapHelper(renderer, modifiers).snap(pixelCoords);
-
-    mNewMapObjectItem->mapObject()->setPosition(pixelCoords);
-    mNewMapObjectItem->syncWithMapObject();
-    mNewMapObjectItem->setZValue(10000); // sync may change it
-    mNewMapObjectItem->setOpacity(0.75);
-}
-
-void CreateTemplateTool::mousePressedWhileCreatingObject(QGraphicsSceneMouseEvent *event)
-{
-    if (event->button() == Qt::RightButton)
-        cancelNewMapObject();
-}
-
-void CreateTemplateTool::mouseReleasedWhileCreatingObject(QGraphicsSceneMouseEvent *event)
-{
-    if (event->button() == Qt::LeftButton)
-        finishNewMapObject();
-}
-
-bool CreateTemplateTool::startNewMapObject(const QPointF &pos, ObjectGroup *objectGroup)
-{
-    if (!CreateObjectTool::startNewMapObject(pos, objectGroup))
-        return false;
-
-    mNewMapObjectItem->setOpacity(0.75);
-    return true;
-}
-
 void CreateTemplateTool::languageChanged()
 {
     CreateObjectTool::languageChanged();
@@ -86,17 +51,17 @@ void CreateTemplateTool::languageChanged()
 void CreateTemplateTool::languageChangedImpl()
 {
     setName(tr("Insert Template"));
-    setShortcut(QKeySequence(tr("V")));
 }
 
 MapObject *CreateTemplateTool::createNewMapObject()
 {
     auto newObjectTemplate = objectTemplate();
-
     if (!newObjectTemplate)
         return nullptr;
+    if (!mapDocument()->templateAllowed(newObjectTemplate))
+        return nullptr;
 
-    MapObject *newMapObject = new MapObject();
+    MapObject *newMapObject = new MapObject;
     newMapObject->setObjectTemplate(newObjectTemplate);
     newMapObject->syncWithTemplate();
     return newMapObject;
