@@ -28,14 +28,14 @@
 #include "utils.h"
 
 namespace Tiled {
-namespace Internal {
 
 CreateTextObjectTool::CreateTextObjectTool(QObject *parent)
     : CreateObjectTool(parent)
 {
-    QIcon icon(QLatin1String(":images/24x24/insert-text.png"));
-    icon.addFile(QLatin1String(":images/48x48/insert-text.png"));
+    QIcon icon(QLatin1String(":images/24/insert-text.png"));
+    icon.addFile(QLatin1String(":images/48/insert-text.png"));
     setIcon(icon);
+    setShortcut(Qt::Key_E);
     Utils::setThemeIcon(this, "insert-text");
     languageChangedImpl();
 }
@@ -52,19 +52,6 @@ void CreateTextObjectTool::mouseMovedWhileCreatingObject(const QPointF &pos, Qt:
 
     mNewMapObjectItem->mapObject()->setPosition(pixelCoords);
     mNewMapObjectItem->syncWithMapObject();
-    mNewMapObjectItem->setZValue(10000); // sync may change it
-}
-
-void CreateTextObjectTool::mousePressedWhileCreatingObject(QGraphicsSceneMouseEvent *event)
-{
-    if (event->button() == Qt::RightButton)
-        cancelNewMapObject();
-}
-
-void CreateTextObjectTool::mouseReleasedWhileCreatingObject(QGraphicsSceneMouseEvent *event)
-{
-    if (event->button() == Qt::LeftButton)
-        finishNewMapObject();
 }
 
 void CreateTextObjectTool::languageChanged()
@@ -76,7 +63,6 @@ void CreateTextObjectTool::languageChanged()
 void CreateTextObjectTool::languageChangedImpl()
 {
     setName(tr("Insert Text"));
-    setShortcut(QKeySequence(tr("E")));
 }
 
 MapObject *CreateTextObjectTool::createNewMapObject()
@@ -91,5 +77,4 @@ MapObject *CreateTextObjectTool::createNewMapObject()
     return newMapObject;
 }
 
-} // namespace Internal
 } // namespace Tiled
