@@ -22,147 +22,53 @@
 
 #include "addremovemapobject.h"
 #include "changepolygon.h"
+#include "createpolygonobjecttool.h"
 #include "geometry.h"
 #include "layer.h"
 #include "map.h"
 #include "mapdocument.h"
 #include "mapobject.h"
 #include "mapobjectitem.h"
-#include "mapobjectmodel.h"
 #include "maprenderer.h"
 #include "mapscene.h"
 #include "objectgroup.h"
+#include "objectselectiontool.h"
+#include "pointhandle.h"
 #include "rangeset.h"
 #include "selectionrectangle.h"
 #include "snaphelper.h"
+#include "toolmanager.h"
 #include "utils.h"
 
 #include <QApplication>
-#include <QGraphicsItem>
 #include <QGraphicsView>
+#include <QKeyEvent>
 #include <QMenu>
-#include <QPainter>
-#include <QPalette>
 #include <QUndoStack>
 
+#include "changeevents.h"
 #include "qtcompat_p.h"
 
 #include <cstdlib>
 
 using namespace Tiled;
-using namespace Tiled::Internal;
-
-namespace Tiled {
-namespace Internal {
-
-/**
- * A handle that allows moving around a point of a polygon.
- */
-class PointHandle : public QGraphicsItem
-{
-public:
-    PointHandle(MapObject *mapObject, int pointIndex)
-        : QGraphicsItem()
-        , mMapObject(mapObject)
-        , mPointIndex(pointIndex)
-        , mSelected(false)
-        , mUnderMouse(false)
-    {
-        setFlags(QGraphicsItem::ItemIgnoresTransformations |
-                 QGraphicsItem::ItemIgnoresParentOpacity);
-        setZValue(10000);
-    }
-
-    enum { Type = UserType + 2 };
-    int type() const override { return Type; }
-
-    MapObject *mapObject() const { return mMapObject; }
-
-    int pointIndex() const { return mPointIndex; }
-
-    // These hide the QGraphicsItem members
-    void setSelected(bool selected);
-    bool isSelected() const { return mSelected; }
-
-    void setUnderMouse(bool underMouse);
-
-    QRectF boundingRect() const override;
-    void paint(QPainter *painter,
-               const QStyleOptionGraphicsItem *option,
-               QWidget *widget = nullptr) override;
-
-private:
-    MapObject *mMapObject;
-    int mPointIndex;
-    bool mSelected;
-    bool mUnderMouse;
-};
-
-} // namespace Internal
-} // namespace Tiled
-
-void PointHandle::setSelected(bool selected)
-{
-    if (mSelected != selected) {
-        mSelected = selected;
-        update();
-    }
-}
-
-void PointHandle::setUnderMouse(bool underMouse)
-{
-    if (mUnderMouse != underMouse) {
-        mUnderMouse = underMouse;
-        update();
-    }
-}
-
-QRectF PointHandle::boundingRect() const
-{
-    return Utils::dpiScaled(QRectF(-7, -7, 14, 14));
-}
-
-void PointHandle::paint(QPainter *painter,
-                        const QStyleOptionGraphicsItem *,
-                        QWidget *)
-{
-    QPen pen(Qt::black);
-    QColor brush(Qt::lightGray);
-
-    if (mSelected)
-        brush = QApplication::palette().highlight().color();
-    if (mUnderMouse)
-        brush = brush.lighter();
-
-    painter->scale(Utils::defaultDpiScale(), Utils::defaultDpiScale());
-    painter->setRenderHint(QPainter::Antialiasing);
-    painter->setPen(pen);
-    painter->setBrush(brush);
-
-    if (mSelected)
-        painter->drawEllipse(QRectF(-5, -5, 10, 10));
-    else
-        painter->drawEllipse(QRectF(-4, -4, 8, 8));
-}
-
 
 EditPolygonTool::EditPolygonTool(QObject *parent)
     : AbstractObjectTool(tr("Edit Polygons"),
-          QIcon(QLatin1String(":images/24x24/tool-edit-polygons.png")),
-          QKeySequence(tr("O")),
+          QIcon(QLatin1String(":images/24/tool-edit-polygons.png")),
+          QKeySequence(Qt::Key_O),
           parent)
     , mSelectionRectangle(new SelectionRectangle)
     , mMousePressed(false)
     , mHoveredHandle(nullptr)
     , mClickedHandle(nullptr)
-    , mClickedObjectItem(nullptr)
-    , mMode(NoMode)
+    , mClickedObject(nullptr)
+    , mAction(NoAction)
 {
 }
 
 EditPolygonTool::~EditPolygonTool()
 {
-    delete mSelectionRectangle;
 }
 
 void EditPolygonTool::activate(MapScene *scene)
@@ -174,36 +80,51 @@ void EditPolygonTool::activate(MapScene *scene)
     // TODO: Could be more optimal by separating the updating of handles from
     // the creation and removal of handles depending on changes in the
     // selection, and by only updating the handles of the objects that changed.
-    connect(mapDocument(), &MapDocument::objectsChanged,
-            this, &EditPolygonTool::updateHandles);
     connect(mapDocument(), &MapDocument::selectedObjectsChanged,
             this, &EditPolygonTool::updateHandles);
-    connect(mapDocument(), &MapDocument::objectsRemoved,
-            this, &EditPolygonTool::objectsRemoved);
 }
 
 void EditPolygonTool::deactivate(MapScene *scene)
 {
-    disconnect(mapDocument(), &MapDocument::objectsChanged,
-               this, &EditPolygonTool::updateHandles);
     disconnect(mapDocument(), &MapDocument::selectedObjectsChanged,
                this, &EditPolygonTool::updateHandles);
-    disconnect(mapDocument(), &MapDocument::objectsRemoved,
-               this, &EditPolygonTool::objectsRemoved);
+
+    abortCurrentAction();
 
     // Delete all handles
-    QMapIterator<MapObject*, QList<PointHandle*> > i(mHandles);
+    QHashIterator<MapObject*, QList<PointHandle*> > i(mHandles);
     while (i.hasNext())
         qDeleteAll(i.next().value());
 
+    mHoveredHandle = nullptr;
+    mHoveredSegment.clear();
     mHandles.clear();
     mSelectedHandles.clear();
-    mHoveredHandle = nullptr;
-    mClickedHandle = nullptr;
+    mHighlightedHandles.clear();
 
     AbstractObjectTool::deactivate(scene);
 }
 
+void EditPolygonTool::keyPressed(QKeyEvent *event)
+{
+    switch (event->key()) {
+    case Qt::Key_Escape:
+        if (mAction != NoAction) {
+            // Abort the current action if any is being performed
+            abortCurrentAction();
+        } else if (!mSelectedHandles.isEmpty()) {
+            // Clear the handle selection if there is one
+            setSelectedHandles(QSet<PointHandle*>());
+        } else {
+            // Switch to object selection tool
+            toolManager()->selectTool(toolManager()->findTool<ObjectSelectionTool>());
+        }
+        return;
+    }
+
+    AbstractObjectTool::keyPressed(event);
+}
+
 void EditPolygonTool::mouseEntered()
 {
 }
@@ -215,43 +136,45 @@ void EditPolygonTool::mouseMoved(const QPointF &pos,
 
     updateHover(pos);
 
-    if (mMode == NoMode && mMousePressed) {
+    if (mAction == NoAction && mMousePressed) {
         QPoint screenPos = QCursor::pos();
         const int dragDistance = (mScreenStart - screenPos).manhattanLength();
-        const bool hasSelection = !mSelectedHandles.isEmpty();
 
         // Use a reduced start drag distance to increase the responsiveness
         if (dragDistance >= QApplication::startDragDistance() / 2) {
             // Holding Alt forces moving current selection
+            const bool forceMove = (modifiers & Qt::AltModifier) && !mSelectedHandles.isEmpty();
+
             // Holding Shift forces selection rectangle
-            if ((mClickedHandle || ((modifiers & Qt::AltModifier) && hasSelection)) &&
-                    !(modifiers & Qt::ShiftModifier))
-                startMoving(pos, modifiers);
-            else
+            const bool forceSelect = modifiers & Qt::ShiftModifier;
+
+            if (!forceSelect && (forceMove || mClickedHandle || mClickedSegment)) {
+                // Move only the clicked handles, if any were not part of the selection
+                if (!forceMove) {
+                    QSet<PointHandle*> handles = clickedHandles();
+                    if (!mSelectedHandles.contains(handles))
+                        setSelectedHandles(handles);
+                }
+
+                startMoving(pos);
+            } else {
                 startSelecting();
+            }
         }
     }
 
-    switch (mMode) {
+    switch (mAction) {
     case Selecting:
         mSelectionRectangle->setRectangle(QRectF(mStart, pos).normalized());
         break;
     case Moving:
         updateMovingItems(pos, modifiers);
         break;
-    case NoMode:
+    case NoAction:
         break;
     }
-}
 
-template <class T>
-static T *first(const QList<QGraphicsItem *> &items)
-{
-    for (QGraphicsItem *item : items) {
-        if (T *t = qgraphicsitem_cast<T*>(item))
-            return t;
-    }
-    return nullptr;
+    mLastMousePos = pos;
 }
 
 static QTransform viewTransform(QGraphicsSceneMouseEvent *event)
@@ -264,9 +187,15 @@ static QTransform viewTransform(QGraphicsSceneMouseEvent *event)
 
 void EditPolygonTool::mousePressed(QGraphicsSceneMouseEvent *event)
 {
-    if (mMode != NoMode) // Ignore additional presses during select/move
+    if (mAction != NoAction) // Ignore additional presses during select/move
         return;
 
+    // Scene or view may have changed since last mouse event
+    updateHover(event->scenePos(), event);
+
+    mClickedHandle = mHoveredHandle;
+    mClickedSegment = mHoveredSegment;
+
     switch (event->button()) {
     case Qt::LeftButton: {
         mMousePressed = true;
@@ -278,28 +207,26 @@ void EditPolygonTool::mousePressed(QGraphicsSceneMouseEvent *event)
                                                                Qt::DescendingOrder,
                                                                viewTransform(event));
 
-        mClickedObjectItem = nullptr;
-        for (int i = 0; i < items.size(); ++i) {
-            if (auto mapObjectItem = qgraphicsitem_cast<MapObjectItem*>(items.at(i))) {
+        mClickedObject = nullptr;
+        for (QGraphicsItem *item : items) {
+            if (!item->isEnabled())
+                continue;
+            if (auto mapObjectItem = qgraphicsitem_cast<MapObjectItem*>(item)) {
                 if (mapObjectItem->mapObject()->objectGroup()->isUnlocked()) {
-                    mClickedObjectItem = mapObjectItem;
+                    mClickedObject = mapObjectItem->mapObject();
                     break;
                 }
             }
         }
-        mClickedHandle = first<PointHandle>(items);
         break;
     }
     case Qt::RightButton: {
-        const QList<QGraphicsItem *> items = mapScene()->items(event->scenePos(),
-                                                               Qt::IntersectsItemShape,
-                                                               Qt::DescendingOrder,
-                                                               viewTransform(event));
+        if (mClickedHandle || mClickedSegment || !mSelectedHandles.isEmpty()) {
+            QSet<PointHandle*> handles = clickedHandles();
+            if (!mSelectedHandles.contains(handles))
+                setSelectedHandles(handles);
 
-        PointHandle *clickedHandle = first<PointHandle>(items);
-        if (clickedHandle || !mSelectedHandles.isEmpty()) {
-            showHandleContextMenu(clickedHandle,
-                                  event->screenPos());
+            showHandleContextMenu(event->screenPos());
         } else {
             AbstractObjectTool::mousePressed(event);
         }
@@ -315,25 +242,27 @@ void EditPolygonTool::mouseReleased(QGraphicsSceneMouseEvent *event)
 {
     if (event->button() != Qt::LeftButton)
         return;
+    if (!mMousePressed)
+        return; // we didn't receive press so we should ignore this release
 
-    switch (mMode) {
-    case NoMode:
-        if (mClickedHandle) {
+    switch (mAction) {
+    case NoAction:
+        if (mClickedHandle || mClickedSegment) {
             QSet<PointHandle*> selection = mSelectedHandles;
+            QSet<PointHandle*> clicked = clickedHandles();
+
             const Qt::KeyboardModifiers modifiers = event->modifiers();
             if (modifiers & (Qt::ShiftModifier | Qt::ControlModifier)) {
-                if (selection.contains(mClickedHandle))
-                    selection.remove(mClickedHandle);
+                if (selection.contains(clicked))
+                    selection.subtract(clicked);
                 else
-                    selection.insert(mClickedHandle);
+                    selection.unite(clicked);
             } else {
-                selection.clear();
-                selection.insert(mClickedHandle);
+                selection = clicked;
             }
             setSelectedHandles(selection);
-        } else if (mClickedObjectItem) {
+        } else if (MapObject *clickedObject = mClickedObject) {
             QList<MapObject*> selection = mapDocument()->selectedObjects();
-            MapObject *clickedObject = mClickedObjectItem->mapObject();
             const Qt::KeyboardModifiers modifiers = event->modifiers();
             if (modifiers & (Qt::ShiftModifier | Qt::ControlModifier)) {
                 int index = selection.indexOf(clickedObject);
@@ -346,20 +275,18 @@ void EditPolygonTool::mouseReleased(QGraphicsSceneMouseEvent *event)
                 selection.append(clickedObject);
             }
             mapDocument()->setSelectedObjects(selection);
-            updateHandles();
         } else if (!mSelectedHandles.isEmpty()) {
             // First clear the handle selection
             setSelectedHandles(QSet<PointHandle*>());
-        } else {
+        } else if (!mapDocument()->selectedObjects().isEmpty()) {
             // If there is no handle selection, clear the object selection
             mapDocument()->setSelectedObjects(QList<MapObject*>());
-            updateHandles();
         }
         break;
     case Selecting:
         updateSelection(event);
-        mapScene()->removeItem(mSelectionRectangle);
-        mMode = NoMode;
+        mapScene()->removeItem(mSelectionRectangle.get());
+        mAction = NoAction;
         break;
     case Moving:
         finishMoving(event->scenePos());
@@ -368,8 +295,36 @@ void EditPolygonTool::mouseReleased(QGraphicsSceneMouseEvent *event)
 
     mMousePressed = false;
     mClickedHandle = nullptr;
+    mClickedSegment.clear();
 
-    updateHover(event->scenePos());
+    updateHover(event->scenePos(), event);
+}
+
+void EditPolygonTool::mouseDoubleClicked(QGraphicsSceneMouseEvent *event)
+{
+    mousePressed(event);
+
+    if (mAction == NoAction && mClickedSegment) {
+        // Split the segment at the location nearest to the mouse
+        QPolygonF oldPolygon = mClickedSegment.object->polygon();
+        QPolygonF newPolygon = oldPolygon;
+        newPolygon.insert(mClickedSegment.index + 1, mClickedSegment.nearestPointOnLine);
+
+        auto splitSegment = new ChangePolygon(mapDocument(),
+                                              mClickedSegment.object,
+                                              newPolygon,
+                                              oldPolygon);
+        splitSegment->setText(tr("Split Segment"));
+
+        mapDocument()->undoStack()->push(splitSegment);
+
+        auto newNodeHandle = mHandles.value(mClickedSegment.object).at(mClickedSegment.index + 1);
+        setSelectedHandle(newNodeHandle);
+        setHighlightedHandles(mSelectedHandles);
+        mHoveredHandle = newNodeHandle;
+        mClickedSegment.clear();
+        mClickedHandle = newNodeHandle;
+    }
 }
 
 void EditPolygonTool::modifiersChanged(Qt::KeyboardModifiers modifiers)
@@ -382,7 +337,6 @@ void EditPolygonTool::languageChanged()
     AbstractObjectTool::languageChanged();
 
     setName(tr("Edit Polygons"));
-    setShortcut(QKeySequence(tr("O")));
 }
 
 void EditPolygonTool::setSelectedHandles(const QSet<PointHandle *> &handles)
@@ -398,6 +352,19 @@ void EditPolygonTool::setSelectedHandles(const QSet<PointHandle *> &handles)
     mSelectedHandles = handles;
 }
 
+void EditPolygonTool::setHighlightedHandles(const QSet<PointHandle *> &handles)
+{
+    for (PointHandle *handle : qAsConst(mHighlightedHandles))
+        if (!handles.contains(handle))
+            handle->setHighlighted(false);
+
+    for (PointHandle *handle : handles)
+        if (!mHighlightedHandles.contains(handle))
+            handle->setHighlighted(true);
+
+    mHighlightedHandles = handles;
+}
+
 /**
  * Creates and removes handle instances as necessary to adapt to a new object
  * selection.
@@ -407,24 +374,32 @@ void EditPolygonTool::updateHandles()
     const QList<MapObject*> &selection = mapDocument()->selectedObjects();
 
     auto deleteHandle = [this](PointHandle *handle) {
+        if (mHoveredHandle == handle)
+            mHoveredHandle = nullptr;
+        if (mClickedHandle == handle)
+            mClickedHandle = nullptr;
         if (handle->isSelected())
             mSelectedHandles.remove(handle);
-        if (handle == mHoveredHandle)
-            mHoveredHandle = nullptr;
+        if (handle->isHighlighted())
+            mHighlightedHandles.remove(handle);
         delete handle;
     };
 
     // First destroy the handles for objects that are no longer selected
-    QMutableMapIterator<MapObject*, QList<PointHandle*> > i(mHandles);
+    QMutableHashIterator<MapObject*, QList<PointHandle*> > i(mHandles);
     while (i.hasNext()) {
         i.next();
         if (!selection.contains(i.key())) {
-            for (PointHandle *handle : i.value())
+            for (PointHandle *handle : qAsConst(i.value()))
                 deleteHandle(handle);
 
             i.remove();
         }
     }
+    if (mHoveredSegment && !selection.contains(mHoveredSegment.object))
+        mHoveredSegment.clear();
+    if (mClickedSegment && !selection.contains(mClickedSegment.object))
+        mClickedSegment.clear();
 
     MapRenderer *renderer = mapDocument()->renderer();
 
@@ -447,30 +422,36 @@ void EditPolygonTool::updateHandles()
         while (pointHandles.size() > polygon.size())
             deleteHandle(pointHandles.takeLast());
 
+        if (pointHandles.isEmpty())
+            continue;
+
+        QPointF objectScreenPos = renderer->pixelToScreenCoords(object->position());
+        QTransform rotate = rotateAt(objectScreenPos, object->rotation());
+        QPointF totalOffset = object->objectGroup()->totalOffset();
+
         // Update the position of all handles
         for (int i = 0; i < pointHandles.size(); ++i) {
             QPointF pixelPos = polygon.at(i) + object->position();
             QPointF screenPos = renderer->pixelToScreenCoords(pixelPos);
-
-            QPointF objectScreenPos = renderer->pixelToScreenCoords(object->position());
-            QTransform rotate = rotateAt(objectScreenPos, object->rotation());
             screenPos = rotate.map(screenPos);
-
-            QPointF totalOffset = object->objectGroup()->totalOffset();
             pointHandles.at(i)->setPos(totalOffset + screenPos);
         }
     }
 }
 
-void EditPolygonTool::objectsRemoved(const QList<MapObject *> &objects)
+void EditPolygonTool::objectsAboutToBeRemoved(const QList<MapObject *> &objects)
 {
-    if (mMode == Moving) {
+    if (mAction == Moving) {
         // Make sure we're not going to try to still change these objects when
         // finishing the move operation.
         // TODO: In addition to avoiding crashes, it would also be good to
         // disallow other actions while moving.
-        for (MapObject *object : objects)
-            mOldPolygons.remove(object);
+        for (MapObject *object : objects) {
+            if (mOldPolygons.contains(object)) {
+                abortCurrentAction(objects);
+                break;
+            }
+        }
     }
 }
 
@@ -492,13 +473,14 @@ void EditPolygonTool::updateSelection(QGraphicsSceneMouseEvent *event)
         QList<MapObject*> selectedObjects;
 
         for (QGraphicsItem *item : intersectedItems) {
+            if (!item->isEnabled())
+                continue;
             auto mapObjectItem = qgraphicsitem_cast<MapObjectItem*>(item);
             if (mapObjectItem && mapObjectItem->mapObject()->objectGroup()->isUnlocked())
                 selectedObjects.append(mapObjectItem->mapObject());
         }
 
         mapDocument()->setSelectedObjects(selectedObjects);
-        updateHandles();
     } else {
         // Update the selected handles
         QSet<PointHandle*> selectedHandles;
@@ -517,19 +499,13 @@ void EditPolygonTool::updateSelection(QGraphicsSceneMouseEvent *event)
 
 void EditPolygonTool::startSelecting()
 {
-    mMode = Selecting;
-    mapScene()->addItem(mSelectionRectangle);
+    mAction = Selecting;
+    mapScene()->addItem(mSelectionRectangle.get());
 }
 
-void EditPolygonTool::startMoving(const QPointF &pos,
-                                  Qt::KeyboardModifiers modifiers)
+void EditPolygonTool::startMoving(const QPointF &pos)
 {
-    // Move only the clicked handle, if it was not part of the selection
-    if (mClickedHandle && !(modifiers & Qt::AltModifier))
-        if (!mSelectedHandles.contains(mClickedHandle))
-            setSelectedHandle(mClickedHandle);
-
-    mMode = Moving;
+    mAction = Moving;
     mStart = pos;
 
     MapRenderer *renderer = mapDocument()->renderer();
@@ -587,7 +563,8 @@ void EditPolygonTool::updateMovingItems(const QPointF &pos,
         // update the polygon
         QPolygonF polygon = object->polygon();
         polygon[handle->pointIndex()] = newPixelPos - object->position();
-        mapDocument()->mapObjectModel()->setObjectPolygon(object, polygon);
+        object->setPolygon(polygon);
+        emit mapDocument()->changed(MapObjectsChangeEvent(object, MapObject::ShapeProperty));
 
         ++i;
     }
@@ -595,8 +572,8 @@ void EditPolygonTool::updateMovingItems(const QPointF &pos,
 
 void EditPolygonTool::finishMoving(const QPointF &pos)
 {
-    Q_ASSERT(mMode == Moving);
-    mMode = NoMode;
+    Q_ASSERT(mAction == Moving);
+    mAction = NoAction;
 
     if (mStart == pos || mOldPolygons.isEmpty()) // Move is a no-op
         return;
@@ -606,7 +583,7 @@ void EditPolygonTool::finishMoving(const QPointF &pos)
 
     // TODO: This isn't really optimal. Would be better to have a single undo
     // command that supports changing multiple map objects.
-    QMapIterator<MapObject*, QPolygonF> i(mOldPolygons);
+    QHashIterator<MapObject*, QPolygonF> i(mOldPolygons);
     while (i.hasNext()) {
         i.next();
         undoStack->push(new ChangePolygon(mapDocument(), i.key(), i.value()));
@@ -618,16 +595,48 @@ void EditPolygonTool::finishMoving(const QPointF &pos)
     mOldPolygons.clear();
 }
 
-void EditPolygonTool::showHandleContextMenu(PointHandle *clickedHandle,
-                                            QPoint screenPos)
+void EditPolygonTool::abortCurrentAction(const QList<MapObject *> &removedObjects)
 {
-    if (clickedHandle && !mSelectedHandles.contains(clickedHandle))
-        setSelectedHandle(clickedHandle);
+    switch (mAction) {
+    case NoAction:
+        break;
+    case Selecting:
+        mapScene()->removeItem(mSelectionRectangle.get());
+        break;
+    case Moving:
+        // Reset the polygons
+        QHashIterator<MapObject*, QPolygonF> i(mOldPolygons);
+        while (i.hasNext()) {
+            i.next();
 
+            MapObject *object = i.key();
+            const QPolygonF &oldPolygon = i.value();
+
+            object->setPolygon(oldPolygon);
+
+            if (!removedObjects.contains(object))
+                emit mapDocument()->changed(MapObjectsChangeEvent(object, MapObject::ShapeProperty));
+        }
+
+        mOldPolygons.clear();
+        break;
+    }
+
+    mAction = NoAction;
+    mMousePressed = false;
+    mClickedHandle = nullptr;
+    mClickedSegment.clear();
+    mClickedObject = nullptr;
+
+    updateHover(mLastMousePos);
+}
+
+void EditPolygonTool::showHandleContextMenu(QPoint screenPos)
+{
     const int n = mSelectedHandles.size();
     Q_ASSERT(n > 0);
 
-    QIcon delIcon(QLatin1String(":images/16x16/edit-delete.png"));
+    QIcon delIcon(QLatin1String(":images/16/edit-delete.png"));
     QString delText = tr("Delete %n Node(s)", "", n);
 
     QMenu menu;
@@ -642,12 +651,12 @@ void EditPolygonTool::showHandleContextMenu(PointHandle *clickedHandle,
     joinNodesAction->setEnabled(n > 1);
     splitSegmentsAction->setEnabled(n > 1);
 
+    const PointHandle *firstHandle = *mSelectedHandles.constBegin();
+    const MapObject *mapObject = firstHandle->mapObject();
+
     bool canDeleteSegment = false;
     if (n == 2) {
-        const PointHandle *firstHandle = *mSelectedHandles.begin();
-        const PointHandle *secondHandle = *(mSelectedHandles.begin() + 1);
-
-        const MapObject *mapObject = firstHandle->mapObject();
+        const PointHandle *secondHandle = *(mSelectedHandles.constBegin() + 1);
         const MapObject *secondMapObject = secondHandle->mapObject();
 
         int indexDifference = std::abs(firstHandle->pointIndex() - secondHandle->pointIndex());
@@ -665,10 +674,39 @@ void EditPolygonTool::showHandleContextMenu(PointHandle *clickedHandle,
     connect(splitSegmentsAction, &QAction::triggered, this, &EditPolygonTool::splitSegments);
     connect(deleteSegment, &QAction::triggered, this, &EditPolygonTool::deleteSegment);
 
+    if (mapObject->shape() == MapObject::Polyline && toolManager()->findTool<CreatePolygonObjectTool>()) {
+        QAction *extendPolyline = menu.addAction(tr("Extend Polyline"));
+
+        bool handleCanBeExtended = (firstHandle->pointIndex() == 0)
+                                   || (firstHandle->pointIndex() == mapObject->polygon().size() - 1);
+
+        extendPolyline->setEnabled(n == 1 && handleCanBeExtended);
+        connect(extendPolyline, &QAction::triggered, this, &EditPolygonTool::extendPolyline);
+    }
+
     menu.exec(screenPos);
 }
 
-typedef QMap<MapObject*, RangeSet<int> > PointIndexesByObject;
+/**
+ * Returns all clicked handles. This may return two handles when a polygon
+ * segment was clicked.
+ */
+QSet<PointHandle *> EditPolygonTool::clickedHandles() const
+{
+    QSet<PointHandle*> handles;
+
+    if (mClickedHandle) {
+        handles.insert(mClickedHandle);
+    } else if (mClickedSegment) {
+        auto handlesForObject = mHandles.value(mClickedSegment.object);
+        handles.insert(handlesForObject.at(mClickedSegment.index));
+        handles.insert(handlesForObject.at((mClickedSegment.index + 1) % handlesForObject.size()));
+    }
+
+    return handles;
+}
+
+typedef QHash<MapObject*, RangeSet<int> > PointIndexesByObject;
 static PointIndexesByObject
 groupIndexesByObject(const QSet<PointHandle*> &handles)
 {
@@ -689,7 +727,7 @@ void EditPolygonTool::deleteNodes()
         return;
 
     PointIndexesByObject p = groupIndexesByObject(mSelectedHandles);
-    QMapIterator<MapObject*, RangeSet<int> > i(p);
+    QHashIterator<MapObject*, RangeSet<int> > i(p);
 
     QUndoStack *undoStack = mapDocument()->undoStack();
 
@@ -714,7 +752,7 @@ void EditPolygonTool::deleteNodes()
 
         if (newPolygon.size() < 2) {
             // We've removed the entire object
-            undoStack->push(new RemoveMapObject(mapDocument(), object));
+            undoStack->push(new RemoveMapObjects(mapDocument(), object));
         } else {
             undoStack->push(new ChangePolygon(mapDocument(), object,
                                               newPolygon,
@@ -725,6 +763,37 @@ void EditPolygonTool::deleteNodes()
     undoStack->endMacro();
 }
 
+void EditPolygonTool::changeEvent(const ChangeEvent &event)
+{
+    AbstractObjectTool::changeEvent(event);
+
+    if (!mapScene())
+        return;
+
+    switch (event.type) {
+    case ChangeEvent::LayerChanged:
+        if (static_cast<const LayerChangeEvent&>(event).properties & LayerChangeEvent::OffsetProperty)
+            updateHandles();
+        break;
+    case ChangeEvent::MapObjectsChanged: {
+        constexpr auto propertiesAffectingHandles =
+                MapObject::PositionProperty |
+                MapObject::RotationProperty |
+                MapObject::ShapeProperty;
+
+        if (static_cast<const MapObjectsChangeEvent&>(event).properties & propertiesAffectingHandles)
+            updateHandles();
+
+        break;
+    }
+    case ChangeEvent::MapObjectsAboutToBeRemoved:
+        objectsAboutToBeRemoved(static_cast<const MapObjectsEvent&>(event).mapObjects);
+        break;
+    default:
+        break;
+    }
+}
+
 /**
  * Joins the nodes at the given \a indexRanges. Each consecutive sequence
  * of nodes will be joined into a single node at the average location.
@@ -852,7 +921,7 @@ void EditPolygonTool::joinNodes()
         return;
 
     const PointIndexesByObject p = groupIndexesByObject(mSelectedHandles);
-    QMapIterator<MapObject*, RangeSet<int> > i(p);
+    QHashIterator<MapObject*, RangeSet<int> > i(p);
 
     QUndoStack *undoStack = mapDocument()->undoStack();
     bool macroStarted = false;
@@ -888,7 +957,7 @@ void EditPolygonTool::splitSegments()
         return;
 
     const PointIndexesByObject p = groupIndexesByObject(mSelectedHandles);
-    QMapIterator<MapObject*, RangeSet<int> > i(p);
+    QHashIterator<MapObject*, RangeSet<int> > i(p);
 
     QUndoStack *undoStack = mapDocument()->undoStack();
     bool macroStarted = false;
@@ -918,6 +987,18 @@ void EditPolygonTool::splitSegments()
         undoStack->endMacro();
 }
 
+void EditPolygonTool::extendPolyline()
+{
+    // Handle is going to be deleted when switching tools
+    PointHandle *selectedHandle = *mSelectedHandles.constBegin();
+    MapObject *mapObject = selectedHandle->mapObject();
+    bool extendingFirst = selectedHandle->pointIndex() == 0;
+
+    auto *polygonObjectsTool = toolManager()->findTool<CreatePolygonObjectTool>();
+    if (toolManager()->selectTool(polygonObjectsTool))
+        polygonObjectsTool->extend(mapObject, extendingFirst);
+}
+
 void EditPolygonTool::deleteSegment()
 {
     if (mSelectedHandles.size() != 2)
@@ -966,30 +1047,99 @@ void EditPolygonTool::deleteSegment()
     }
 }
 
-void EditPolygonTool::updateHover(const QPointF &pos)
+/**
+ * Returns the shortest distance between \a point and \a line.
+ */
+static qreal distanceOfPointToLine(const QLineF &line, QPointF point, QPointF &nearestPointOnLine)
+{
+    // implementation is based on QLineF::intersect
+    const QPointF d = line.p2() - line.p1();
+    const qreal denominator = d.x() * d.x() + d.y() * d.y();
+    if (denominator == 0) {
+        nearestPointOnLine = line.p1();
+        return QLineF(point, line.p1()).length();
+    }
+
+    const QPointF c = point - line.p1();
+    const qreal na = qBound<qreal>(0, (d.x() * c.x() + d.y() * c.y()) / denominator, 1);
+
+    nearestPointOnLine = line.p1() + d * na;
+    return QLineF(point, nearestPointOnLine).length();
+}
+
+void EditPolygonTool::updateHover(const QPointF &scenePos, QGraphicsSceneMouseEvent *event)
 {
     PointHandle *hoveredHandle = nullptr;
+    InteractedSegment hoveredSegment;
 
-    switch (mMode) {
+    switch (mAction) {
     case Moving:    // while moving, optionally keep clicked handle hovered
         if (mClickedHandle && mClickedHandle->isSelected())
             hoveredHandle = mClickedHandle;
         break;
-    case NoMode:
-        if (QGraphicsView *view = mapScene()->views().first()) {
-            QGraphicsItem *hoveredItem = mapScene()->itemAt(pos, view->transform());
-            hoveredHandle = qgraphicsitem_cast<PointHandle*>(hoveredItem);
+    case NoAction: {
+        QTransform transform;
+        if (event)
+            transform = viewTransform(event);
+        else if (QGraphicsView *view = mapScene()->views().first())
+            transform = view->transform();
+
+        QGraphicsItem *hoveredItem = mapScene()->itemAt(scenePos, transform);
+        hoveredHandle = qgraphicsitem_cast<PointHandle*>(hoveredItem);
+
+        if (!hoveredHandle) {
+            // check if we're hovering a line segment
+            MapRenderer *renderer = mapDocument()->renderer();
+
+            const qreal hoverDistance = 7 / renderer->painterScale();
+            qreal minDistance = std::numeric_limits<qreal>::max();
+
+            for (MapObject *object : mapDocument()->selectedObjects()) {
+                if (object->shape() != MapObject::Polygon && object->shape() != MapObject::Polyline)
+                    continue;
+
+                // Translate mouse position to local pixel coordinates...
+                const QPointF totalOffset = object->objectGroup()->totalOffset();
+                const QPointF objectScreenPos = renderer->pixelToScreenCoords(object->position());
+                const QTransform rotate = rotateAt(objectScreenPos, -object->rotation());
+                const QPointF rotatedMouseScenePos = rotate.map(scenePos - totalOffset);
+                const QPointF mousePixelCoords = renderer->screenToPixelCoords(rotatedMouseScenePos);
+                const QPointF localMousePixelCoords = mousePixelCoords - object->position();
+
+                const QPolygonF &polygon = object->polygon();
+                const int end = object->shape() == MapObject::Polygon ? polygon.size(): polygon.size() - 1;
+
+                for (int i = 0; i < end; ++i) {
+                    const QLineF line(polygon.at(i), polygon.at((i + 1) % polygon.size()));
+                    QPointF nearestPointOnLine;
+                    const qreal distance = distanceOfPointToLine(line, localMousePixelCoords, nearestPointOnLine);
+                    if (distance < hoverDistance && distance < minDistance) {
+                        minDistance = distance;
+                        hoveredSegment.object = object;
+                        hoveredSegment.index = i;
+                        hoveredSegment.nearestPointOnLine = nearestPointOnLine;
+                    }
+                }
+            }
         }
         break;
+    }
     case Selecting:
         break;      // no hover while selecting
     }
 
-    if (mHoveredHandle != hoveredHandle) {
-        if (mHoveredHandle)
-            mHoveredHandle->setUnderMouse(false);
-        if (hoveredHandle)
-            hoveredHandle->setUnderMouse(true);
-        mHoveredHandle = hoveredHandle;
+    QSet<PointHandle*> highlightedHandles;
+
+    if (hoveredHandle) {
+        highlightedHandles.insert(hoveredHandle);
+    } else if (hoveredSegment) {
+        auto handles = mHandles.value(hoveredSegment.object);
+        highlightedHandles.insert(handles.at(hoveredSegment.index));
+        highlightedHandles.insert(handles.at((hoveredSegment.index + 1) % handles.size()));
     }
+
+    setHighlightedHandles(highlightedHandles);
+
+    mHoveredHandle = hoveredHandle;
+    mHoveredSegment = hoveredSegment;
 }
