@@ -24,6 +24,7 @@
 
 #include "document.h"
 #include "layer.h"
+#include "map.h"
 #include "mapformat.h"
 #include "tiled.h"
 #include "tileset.h"
@@ -32,6 +33,8 @@
 #include <QPointer>
 #include <QRegion>
 
+#include <memory>
+
 class QModelIndex;
 class QPoint;
 class QRect;
@@ -48,11 +51,13 @@ class Terrain;
 class Tile;
 class WangSet;
 
-namespace Internal {
-
 class LayerModel;
+class MapDocument;
 class MapObjectModel;
 class TileSelectionModel;
+class EditableMap;
+
+using MapDocumentPtr = QSharedPointer<MapDocument>;
 
 /**
  * Represents an editable map. The purpose of this class is to make sure that
@@ -67,6 +72,8 @@ class MapDocument : public Document
 {
     Q_OBJECT
 
+    Q_PROPERTY(Map *map READ map CONSTANT)
+
 public:
     enum TileLayerChangeFlag {
         LayerDrawMarginsChanged,
@@ -76,22 +83,23 @@ public:
     Q_FLAG(TileLayerChangeFlags)
 
     /**
-     * Constructs a map document around the given map. The map document takes
-     * ownership of the map.
+     * Constructs a map document around the given map.
      */
-    MapDocument(Map *map, const QString &fileName = QString());
+    MapDocument(std::unique_ptr<Map> map);
 
     ~MapDocument() override;
 
+    MapDocumentPtr sharedFromThis() { return qSharedPointerCast<MapDocument>(Document::sharedFromThis()); }
+
     bool save(const QString &fileName, QString *error = nullptr) override;
 
     /**
      * Loads a map and returns a MapDocument instance on success. Returns null
      * on error and sets the \a error message.
      */
-    static MapDocument *load(const QString &fileName,
-                             MapFormat *format,
-                             QString *error = nullptr);
+    static MapDocumentPtr load(const QString &fileName,
+                               MapFormat *format,
+                               QString *error = nullptr);
 
     MapFormat *readerFormat() const;
     void setReaderFormat(MapFormat *format);
@@ -99,6 +107,9 @@ public:
     FileFormat *writerFormat() const override;
     void setWriterFormat(MapFormat *format);
 
+    QString lastExportFileName() const override;
+    void setLastExportFileName(const QString &fileName) override;
+
     MapFormat *exportFormat() const override;
     void setExportFormat(FileFormat *format) override;
 
@@ -108,7 +119,9 @@ public:
      * Returns the map instance. Be aware that directly modifying the map will
      * not allow the GUI to update itself appropriately.
      */
-    Map *map() const { return mMap; }
+    Map *map() const { return mMap.get(); }
+
+    Tiled::EditableAsset *editable() override;
 
     int layerIndex(const Layer *layer) const;
 
@@ -119,12 +132,18 @@ public:
     Layer *currentLayer() const { return mCurrentLayer; }
     void setCurrentLayer(Layer *layer);
 
+    const QList<Layer*> &selectedLayers() const { return mSelectedLayers; }
+    void setSelectedLayers(const QList<Layer*> &layers);
+
+    void switchCurrentLayer(Layer *layer);
+    void switchSelectedLayers(const QList<Layer*> &layers);
+
     /**
      * Resize this map to the given \a size, while at the same time shifting
      * the contents by \a offset. If \a removeObjects is true then all objects
      * which are outside the map will be removed.
      */
-    void resizeMap(const QSize &size, const QPoint &offset, bool removeObjects);
+    void resizeMap(QSize size, QPoint offset, bool removeObjects);
 
     void autocropMap();
 
@@ -133,7 +152,7 @@ public:
      * wraps on the X or Y axis.
      */
     void offsetMap(const QList<Layer *> &layers,
-                   const QPoint &offset,
+                   QPoint offset,
                    const QRect &bounds,
                    bool wrapX, bool wrapY);
 
@@ -141,19 +160,26 @@ public:
     void rotateSelectedObjects(RotateDirection direction);
 
     Layer *addLayer(Layer::TypeFlag layerType);
-    void groupLayer(Layer *layer);
-    void ungroupLayer(Layer *layer);
-    void duplicateLayer();
-    void mergeLayerDown();
-    void moveLayerUp(Layer *layer);
-    void moveLayerDown(Layer *layer);
-    void removeLayer(Layer *layer);
-    void toggleOtherLayers(Layer *layer);
+    void groupLayers(const QList<Layer *> &layers);
+    void ungroupLayers(const QList<Layer *> &layers);
+    void duplicateLayers(const QList<Layer *> &layers);
+    void mergeLayersDown(const QList<Layer *> &layers);
+    void moveLayersUp(const QList<Layer *> &layers);
+    void moveLayersDown(const QList<Layer *> &layers);
+    void removeLayers(const QList<Layer *> &layers);
+    void toggleLayers(const QList<Layer *> &layers);
+    void toggleLockLayers(const QList<Layer *> &layers);
+    void toggleOtherLayers(const QList<Layer *> &layers);
+    void toggleLockOtherLayers(const QList<Layer *> &layers);
 
     void insertTileset(int index, const SharedTileset &tileset);
     void removeTilesetAt(int index);
     SharedTileset replaceTileset(int index, const SharedTileset &tileset);
 
+    void paintTileLayers(const Map *map, bool mergeable = false,
+                         QVector<SharedTileset> *missingTilesets = nullptr,
+                         QHash<TileLayer *, QRegion> *paintedRegions = nullptr);
+
     void replaceObjectTemplate(const ObjectTemplate *oldObjectTemplate,
                                const ObjectTemplate *newObjectTemplate);
 
@@ -176,7 +202,7 @@ public:
     /**
      * Returns the map renderer.
      */
-    MapRenderer *renderer() const { return mRenderer; }
+    MapRenderer *renderer() const { return mRenderer.get(); }
 
     /**
      * Creates the map renderer. Should be called after changing the map
@@ -210,12 +236,22 @@ public:
 
     QList<Object*> currentObjects() const override;
 
+    MapObject *hoveredMapObject() const { return mHoveredMapObject; }
+    void setHoveredMapObject(MapObject *object);
+
     void unifyTilesets(Map *map);
     void unifyTilesets(Map *map, QVector<SharedTileset> &missingTilesets);
 
     bool allowHidingObjects() const { return mAllowHidingObjects; }
     void setAllowHidingObjects(bool value) { mAllowHidingObjects = value; }
 
+    bool allowTileObjects() const { return mAllowTileObjects; }
+    void setAllowTileObjects(bool value) { mAllowTileObjects = value; }
+
+    bool templateAllowed(const ObjectTemplate *objectTemplate) const;
+
+    void checkIssues() override;
+
 signals:
     /**
      * Emitted when the selected tile region changes. Sends the currently
@@ -224,11 +260,27 @@ signals:
     void selectedAreaChanged(const QRegion &newSelection,
                              const QRegion &oldSelection);
 
+    /**
+     * Emitted when the list of selected layers changes.
+     */
+    void selectedLayersChanged();
+
     /**
      * Emitted when the list of selected objects changes.
      */
     void selectedObjectsChanged();
 
+    /**
+     * Emitted when the hovered object changes. Use \a previous with caution,
+     * because it may reference an object that was removed.
+     */
+    void hoveredMapObjectChanged(MapObject *object, MapObject *previous);
+
+    /**
+     * Emitted when the map view should focus on the given object.
+     */
+    void focusMapObjectRequested(MapObject *object);
+
     /**
      * Emitted when the map size or its tile size changes.
      */
@@ -237,7 +289,6 @@ signals:
     void layerAdded(Layer *layer);
     void layerAboutToBeRemoved(GroupLayer *parentLayer, int index);
     void layerRemoved(Layer *layer);
-    void layerChanged(Layer *layer);
 
     /**
      * Emitted after a new layer was added and the name should be edited.
@@ -265,12 +316,6 @@ signals:
 
     void tileLayerChanged(TileLayer *layer, TileLayerChangeFlags flags);
 
-    /**
-     * Should be emitted when changing the color or drawing order of an object
-     * group.
-     */
-    void objectGroupChanged(ObjectGroup *objectGroup);
-
     /**
      * Should be emitted when changing the image or the transparent color of
      * an image layer.
@@ -286,11 +331,7 @@ signals:
     void objectTemplateReplaced(const ObjectTemplate *newObjectTemplate,
                                 const ObjectTemplate *oldObjectTemplate);
 
-    void objectsAdded(const QList<MapObject*> &objects);
     void objectsInserted(ObjectGroup *objectGroup, int first, int last);
-    void objectsRemoved(const QList<MapObject*> &objects);
-    void objectsChanged(const QList<MapObject*> &objects);
-    void objectsTypeChanged(const QList<MapObject*> &objects);
     void objectsIndexChanged(ObjectGroup *objectGroup, int first, int last);
 
     // emitted from the TilesetDocument
@@ -298,9 +339,16 @@ signals:
     void tilesetTileOffsetChanged(Tileset *tileset);
     void tileTypeChanged(Tile *tile);
     void tileImageSourceChanged(Tile *tile);
+    void tileProbabilityChanged(Tile *tile);
+    void tileObjectGroupChanged(Tile *tile);
 
-private slots:
-    void onObjectsRemoved(const QList<MapObject*> &objects);
+public slots:
+    void updateTemplateInstances(const ObjectTemplate *objectTemplate);
+    void selectAllInstances(const ObjectTemplate *objectTemplate);
+    void deselectObjects(const QList<MapObject*> &objects);
+
+private:
+    void onChanged(const ChangeEvent &change);
 
     void onMapObjectModelRowsInserted(const QModelIndex &parent, int first, int last);
     void onMapObjectModelRowsInsertedOrRemoved(const QModelIndex &parent, int first, int last);
@@ -311,12 +359,6 @@ private slots:
     void onLayerAboutToBeRemoved(GroupLayer *groupLayer, int index);
     void onLayerRemoved(Layer *layer);
 
-public slots:
-    void updateTemplateInstances(const ObjectTemplate *objectTemplate);
-    void selectAllInstances(const ObjectTemplate *objectTemplate);
-
-private:
-    void deselectObjects(const QList<MapObject*> &objects);
     void moveObjectIndex(const MapObject *object, int count);
 
     /*
@@ -325,18 +367,19 @@ private:
      */
     QPointer<MapFormat> mReaderFormat;
     QPointer<MapFormat> mWriterFormat;
-    QPointer<MapFormat> mExportFormat;
-    Map *mMap;
+    std::unique_ptr<Map> mMap;
     LayerModel *mLayerModel;
     QRegion mSelectedArea;
+    QList<Layer*> mSelectedLayers;
     QList<MapObject*> mSelectedObjects;
-    MapRenderer *mRenderer;
+    MapObject *mHoveredMapObject;       /**< Map object with mouse on top. */
+    std::unique_ptr<MapRenderer> mRenderer;
     Layer *mCurrentLayer;
     MapObjectModel *mMapObjectModel;
     bool mAllowHidingObjects = true;
+    bool mAllowTileObjects = true;
 };
 
-} // namespace Internal
 } // namespace Tiled
 
-Q_DECLARE_OPERATORS_FOR_FLAGS(Tiled::Internal::MapDocument::TileLayerChangeFlags)
+Q_DECLARE_OPERATORS_FOR_FLAGS(Tiled::MapDocument::TileLayerChangeFlags)
