@@ -21,7 +21,11 @@
 #include "mapview.h"
 
 #include "flexiblescrollbar.h"
+#include "mapdocument.h"
+#include "mapobject.h"
+#include "maprenderer.h"
 #include "mapscene.h"
+#include "objectgroup.h"
 #include "preferences.h"
 #include "utils.h"
 #include "zoomable.h"
@@ -39,11 +43,9 @@
 #endif
 
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 MapView::MapView(QWidget *parent, Mode mode)
     : QGraphicsView(parent)
-    , mHandScrolling(false)
     , mMode(mode)
     , mZoomable(new Zoomable(this))
 {
@@ -55,7 +57,7 @@ MapView::MapView(QWidget *parent, Mode mode)
 #ifndef QT_NO_OPENGL
     Preferences *prefs = Preferences::instance();
     setUseOpenGL(prefs->useOpenGL());
-    connect(prefs, SIGNAL(useOpenGLChanged(bool)), SLOT(setUseOpenGL(bool)));
+    connect(prefs, &Preferences::useOpenGLChanged, this, &MapView::setUseOpenGL);
 #endif
 
     QWidget *v = viewport();
@@ -89,9 +91,18 @@ MapView::~MapView()
 
 void MapView::setScene(MapScene *scene)
 {
+    if (MapScene *currentScene = mapScene())
+        currentScene->disconnect(this);
+
     QGraphicsView::setScene(scene);
-    if (scene)
+
+    if (scene) {
         updateSceneRect(scene->sceneRect());
+        connect(scene, &MapScene::mapDocumentChanged,
+                this, &MapView::setMapDocument);
+    }
+
+    setMapDocument(scene ? scene->mapDocument() : nullptr);
 }
 
 MapScene *MapView::mapScene() const
@@ -99,6 +110,27 @@ MapScene *MapView::mapScene() const
     return static_cast<MapScene*>(scene());
 }
 
+qreal MapView::scale() const
+{
+    return mZoomable->scale();
+}
+
+void MapView::setScale(qreal scale)
+{
+    mZoomable->setScale(scale);
+}
+
+void MapView::fitMapInView()
+{
+    const QRectF rect = mapScene()->mapBoundingRect();
+    if (rect.isEmpty())
+        return;
+
+    // Scale and center map to fit in view
+    centerOn(rect.center());
+    setScale(std::min(width() / rect.width(), height() / rect.height()) * 0.95);
+}
+
 void MapView::adjustScale(qreal scale)
 {
     const QTransform newTransform = QTransform::fromScale(scale, scale);
@@ -166,6 +198,32 @@ void MapView::updateSceneRect(const QRectF &sceneRect, const QTransform &transfo
     setSceneRect(expandedSceneRect);
 }
 
+void MapView::focusMapObject(MapObject *mapObject)
+{
+    // FIXME: This is not always the visual center
+    const QPointF center = mapObject->bounds().center();
+    const QPointF offset = mapObject->objectGroup()->totalOffset();
+    const QPointF focus = center + offset;
+
+    centerOn(mMapDocument->renderer()->pixelToScreenCoords(focus));
+}
+
+void MapView::setMapDocument(MapDocument *mapDocument)
+{
+    if (mMapDocument == mapDocument)
+        return;
+
+    if (mMapDocument)
+        mMapDocument->disconnect(this);
+
+    mMapDocument = mapDocument;
+
+    if (mapDocument) {
+        connect(mapDocument, &MapDocument::focusMapObjectRequested,
+                this, &MapView::focusMapObject);
+    }
+}
+
 void MapView::setHandScrolling(bool handScrolling)
 {
     if (mHandScrolling == handScrolling)
@@ -194,7 +252,7 @@ void MapView::forceCenterOn(const QPointF &pos)
 {
     // This is only to make it update QGraphicsViewPrivate::lastCenterPoint,
     // just in case this is important.
-    centerOn(pos);
+    QGraphicsView::centerOn(pos);
 
     auto hBar = static_cast<FlexibleScrollBar*>(horizontalScrollBar());
     auto vBar = static_cast<FlexibleScrollBar*>(verticalScrollBar());
@@ -248,6 +306,16 @@ bool MapView::event(QEvent *e)
     return QGraphicsView::event(e);
 }
 
+void MapView::showEvent(QShowEvent *event)
+{
+    if (!mViewInitialized) {
+        fitMapInView();
+        mViewInitialized = true;
+    }
+
+    QGraphicsView::showEvent(event);
+}
+
 void MapView::hideEvent(QHideEvent *event)
 {
     // Disable hand scrolling when the view gets hidden in any way
@@ -402,12 +470,12 @@ void MapView::handlePinchGesture(QPinchGesture *pinch)
     setTransformationAnchor(QGraphicsView::AnchorViewCenter);
 }
 
-void MapView::adjustCenterFromMousePosition(QPoint &mousePos)
+void MapView::adjustCenterFromMousePosition(QPoint mousePos)
 {
     // Place the last known mouse scene pos below the mouse again
     QWidget *view = viewport();
     QPointF viewCenterScenePos = mapToScene(view->rect().center());
     QPointF mouseScenePos = mapToScene(view->mapFromGlobal(mousePos));
     QPointF diff = viewCenterScenePos - mouseScenePos;
-    centerOn(mLastMouseScenePos + diff);
+    QGraphicsView::centerOn(mLastMouseScenePos + diff);
 }
