@@ -38,8 +38,9 @@
 #include <QStyledItemDelegate>
 #include <QToolBar>
 
+#include "qtcompat_p.h"
+
 namespace Tiled {
-namespace Internal {
 
 class ColorDelegate : public QStyledItemDelegate
 {
@@ -123,14 +124,14 @@ ObjectTypesEditor::ObjectTypesEditor(QWidget *parent)
     mRemovePropertyAction->setEnabled(false);
     mRenamePropertyAction->setEnabled(false);
 
-    QIcon addIcon(QLatin1String(":/images/22x22/add.png"));
-    QIcon removeIcon(QLatin1String(":/images/22x22/remove.png"));
+    QIcon addIcon(QLatin1String(":/images/22/add.png"));
+    QIcon removeIcon(QLatin1String(":/images/22/remove.png"));
 
     mAddObjectTypeAction->setIcon(addIcon);
     mRemoveObjectTypeAction->setIcon(removeIcon);
     mAddPropertyAction->setIcon(addIcon);
     mRemovePropertyAction->setIcon(removeIcon);
-    mRenamePropertyAction->setIcon(QIcon(QLatin1String(":/images/16x16/rename.png")));
+    mRenamePropertyAction->setIcon(QIcon(QLatin1String(":/images/16/rename.png")));
 
     Utils::setThemeIcon(mAddObjectTypeAction, "add");
     Utils::setThemeIcon(mRemoveObjectTypeAction, "remove");
@@ -156,27 +157,27 @@ ObjectTypesEditor::ObjectTypesEditor(QWidget *parent)
             this, &ObjectTypesEditor::selectedObjectTypesChanged);
     connect(mObjectTypesModel, &ObjectTypesModel::modelReset,
             this, &ObjectTypesEditor::selectFirstType);
-    connect(mUi->objectTypesTable, SIGNAL(doubleClicked(QModelIndex)),
-            SLOT(objectTypeIndexClicked(QModelIndex)));
+    connect(mUi->objectTypesTable, &QAbstractItemView::doubleClicked,
+            this, &ObjectTypesEditor::objectTypeIndexClicked);
 
-    connect(mAddObjectTypeAction, SIGNAL(triggered()),
-            SLOT(addObjectType()));
-    connect(mRemoveObjectTypeAction, SIGNAL(triggered()),
-            SLOT(removeSelectedObjectTypes()));
+    connect(mAddObjectTypeAction, &QAction::triggered,
+            this, &ObjectTypesEditor::addObjectType);
+    connect(mRemoveObjectTypeAction, &QAction::triggered,
+            this, &ObjectTypesEditor::removeSelectedObjectTypes);
 
-    connect(mAddPropertyAction, SIGNAL(triggered()),
-            SLOT(addProperty()));
-    connect(mRemovePropertyAction, SIGNAL(triggered()),
-            SLOT(removeProperty()));
-    connect(mRenamePropertyAction, SIGNAL(triggered()),
-            SLOT(renameProperty()));
+    connect(mAddPropertyAction, &QAction::triggered,
+            this, &ObjectTypesEditor::openAddPropertyDialog);
+    connect(mRemovePropertyAction, &QAction::triggered,
+            this, &ObjectTypesEditor::removeProperty);
+    connect(mRenamePropertyAction, &QAction::triggered,
+            this, &ObjectTypesEditor::renameProperty);
 
-    connect(mUi->actionChooseFile, SIGNAL(triggered()),
-            SLOT(chooseObjectTypesFile()));
-    connect(mUi->actionImport, SIGNAL(triggered()),
-            SLOT(importObjectTypes()));
-    connect(mUi->actionExport, SIGNAL(triggered()),
-            SLOT(exportObjectTypes()));
+    connect(mUi->actionChooseFile, &QAction::triggered,
+            this, &ObjectTypesEditor::chooseObjectTypesFile);
+    connect(mUi->actionImport, &QAction::triggered,
+            this, &ObjectTypesEditor::importObjectTypes);
+    connect(mUi->actionExport, &QAction::triggered,
+            this, &ObjectTypesEditor::exportObjectTypes);
 
     connect(mObjectTypesModel, &ObjectTypesModel::dataChanged,
             this, &ObjectTypesEditor::applyObjectTypes);
@@ -290,7 +291,7 @@ void ObjectTypesEditor::applyObjectTypes()
     }
 }
 
-void ObjectTypesEditor::applyProperty(const QString &name, const QVariant &value)
+void ObjectTypesEditor::applyPropertyToSelectedTypes(const QString &name, const QVariant &value)
 {
     const auto selectionModel = mUi->objectTypesTable->selectionModel();
     const auto selectedRows = selectionModel->selectedRows();
@@ -304,7 +305,7 @@ void ObjectTypesEditor::applyProperty(const QString &name, const QVariant &value
     applyObjectTypes();
 }
 
-void ObjectTypesEditor::removeProperty(const QString &name)
+void ObjectTypesEditor::removePropertyFromSelectedTypes(const QString &name)
 {
     const auto selectionModel = mUi->objectTypesTable->selectionModel();
     const auto selectedRows = selectionModel->selectedRows();
@@ -370,8 +371,8 @@ void ObjectTypesEditor::importObjectTypes()
 
     if (serializer.readObjectTypes(fileName, objectTypes)) {
         ObjectTypes currentTypes = mObjectTypesModel->objectTypes();
-        for (const ObjectType &type : objectTypes) {
-            auto it = std::find_if(currentTypes.begin(), currentTypes.end(), [&type](ObjectType &existingType) {
+        for (const ObjectType &type : qAsConst(objectTypes)) {
+            auto it = std::find_if(currentTypes.begin(), currentTypes.end(), [&type](const ObjectType &existingType) {
                 return existingType.name == type.name;
             });
 
@@ -465,7 +466,7 @@ void ObjectTypesEditor::propertyValueChanged(QtProperty *property,
     if (!mUi->propertiesView->topLevelItem(property))
         return;
 
-    applyProperty(property->propertyName(), value);
+    applyPropertyToSelectedTypes(property->propertyName(), value);
 }
 
 QtVariantProperty *ObjectTypesEditor::createProperty(int type,
@@ -486,7 +487,7 @@ QtVariantProperty *ObjectTypesEditor::createProperty(int type,
     return property;
 }
 
-void ObjectTypesEditor::addProperty()
+void ObjectTypesEditor::openAddPropertyDialog()
 {
     AddPropertyDialog dialog(window());
     if (dialog.exec() == AddPropertyDialog::Accepted)
@@ -498,12 +499,12 @@ void ObjectTypesEditor::addProperty(const QString &name, const QVariant &value)
     if (name.isEmpty())
         return;
 
-    applyProperty(name, value);
+    applyPropertyToSelectedTypes(name, value);
     updateProperties();
-    editCustomProperty(name);
+    editProperty(name);
 }
 
-void ObjectTypesEditor::editCustomProperty(const QString &name)
+void ObjectTypesEditor::editProperty(const QString &name)
 {
     QtVariantProperty *property = mNameToProperty.value(name);
     if (!property)
@@ -534,7 +535,7 @@ void ObjectTypesEditor::removeProperty()
     mProperties.remove(name);
     delete mNameToProperty.take(name);
 
-    removeProperty(name);
+    removePropertyFromSelectedTypes(name);
 }
 
 void ObjectTypesEditor::renameProperty()
@@ -546,14 +547,16 @@ void ObjectTypesEditor::renameProperty()
     const QString oldName = item->property()->propertyName();
 
     QInputDialog *dialog = new QInputDialog(mUi->propertiesView);
+    dialog->setAttribute(Qt::WA_DeleteOnClose);
     dialog->setInputMode(QInputDialog::TextInput);
     dialog->setLabelText(tr("Name:"));
     dialog->setTextValue(oldName);
     dialog->setWindowTitle(tr("Rename Property"));
-    dialog->open(this, SLOT(renameProperty(QString)));
+    connect(dialog, &QInputDialog::textValueSelected, this, &ObjectTypesEditor::renamePropertyTo);
+    dialog->open();
 }
 
-void ObjectTypesEditor::renameProperty(const QString &name)
+void ObjectTypesEditor::renamePropertyTo(const QString &name)
 {
     if (name.isEmpty())
         return;
@@ -600,5 +603,4 @@ void ObjectTypesEditor::currentItemChanged(QtBrowserItem *item)
     mRenamePropertyAction->setEnabled(itemSelected);
 }
 
-} // namespace Internal
 } // namespace Tiled
