@@ -31,7 +31,6 @@
 class QSettings;
 
 namespace Tiled {
-namespace Internal {
 
 /**
  * This class holds user preferences and provides a convenient interface to
@@ -45,18 +44,20 @@ public:
     static Preferences *instance();
     static void deleteInstance();
 
-    bool showGrid() const { return mShowGrid; }
-    bool showTileObjectOutlines() const { return mShowTileObjectOutlines; }
-    bool showTileAnimations() const { return mShowTileAnimations; }
-    bool snapToGrid() const { return mSnapToGrid; }
-    bool snapToFineGrid() const { return mSnapToFineGrid; }
-    bool snapToPixels() const { return mSnapToPixels; }
-    QColor gridColor() const { return mGridColor; }
-    int gridFine() const { return mGridFine; }
-    qreal objectLineWidth() const { return mObjectLineWidth; }
+    bool showGrid() const;
+    bool showTileObjectOutlines() const;
+    bool showTileAnimations() const;
+    bool showTileCollisionShapes() const;
+    bool snapToGrid() const;
+    bool snapToFineGrid() const;
+    bool snapToPixels() const;
+    QColor gridColor() const;
+    int gridFine() const;
+    qreal objectLineWidth() const;
 
-    bool highlightCurrentLayer() const { return mHighlightCurrentLayer; }
-    bool showTilesetGrid() const { return mShowTilesetGrid; }
+    bool highlightCurrentLayer() const;
+    bool highlightHoveredObject() const;
+    bool showTilesetGrid() const;
 
     enum ObjectLabelVisiblity {
         NoObjectLabels,
@@ -67,6 +68,9 @@ public:
     ObjectLabelVisiblity objectLabelVisibility() const;
     void setObjectLabelVisibility(ObjectLabelVisiblity visiblity);
 
+    bool labelForHoveredObject() const;
+    void setLabelForHoveredObject(bool enabled);
+
     enum ApplicationStyle {
         SystemDefaultStyle,
         FusionStyle,
@@ -88,19 +92,31 @@ public:
     Map::RenderOrder mapRenderOrder() const;
     void setMapRenderOrder(Map::RenderOrder mapRenderOrder);
 
-    bool dtdEnabled() const;
-    void setDtdEnabled(bool enabled);
-
     bool safeSavingEnabled() const;
     void setSafeSavingEnabled(bool enabled);
 
+    bool exportOnSave() const;
+    void setExportOnSave(bool enabled);
+
+    enum ExportOption {
+        EmbedTilesets                   = 0x1,
+        DetachTemplateInstances         = 0x2,
+        ResolveObjectTypesAndProperties = 0x4,
+        ExportMinimized                 = 0x8,
+    };
+    Q_DECLARE_FLAGS(ExportOptions, ExportOption)
+
+    ExportOptions exportOptions() const;
+    void setExportOption(ExportOption option, bool value);
+    bool exportOption(ExportOption option) const;
+
     QString language() const;
     void setLanguage(const QString &language);
 
     bool reloadTilesetsOnChange() const;
     void setReloadTilesetsOnChanged(bool value);
 
-    bool useOpenGL() const { return mUseOpenGL; }
+    bool useOpenGL() const;
     void setUseOpenGL(bool useOpenGL);
 
     void setObjectTypes(const ObjectTypes &objectTypes);
@@ -110,13 +126,14 @@ public:
         ObjectTemplateFile,
         ImageFile,
         ExportedFile,
-        ExternalTileset
+        ExternalTileset,
+        WorldFile
     };
 
     QString lastPath(FileType fileType) const;
     void setLastPath(FileType fileType, const QString &path);
 
-    bool automappingDrawing() const { return mAutoMapDrawing; }
+    bool automappingDrawing() const;
 
     QString mapsDirectory() const;
     void setMapsDirectory(const QString &path);
@@ -136,10 +153,10 @@ public:
     bool isPatron() const;
     void setPatron(bool isPatron);
 
-    bool shouldShowPatreonDialog() const;
-    void setPatreonDialogReminder(const QDate &date);
+    bool shouldShowDonationDialog() const;
+    void setDonationDialogReminder(const QDate &date);
 
-    enum { MaxRecentFiles = 8 };
+    enum { MaxRecentFiles = 12 };
     QStringList recentFiles() const;
     QString fileDialogStartLocation() const;
     void addRecentFile(const QString &fileName);
@@ -149,18 +166,22 @@ public:
     bool checkForUpdates() const;
     void setCheckForUpdates(bool on);
 
+    bool displayNews() const;
+    void setDisplayNews(bool on);
+
     bool wheelZoomsByDefault() const;
 
     /**
      * Provides access to the QSettings instance to allow storing/retrieving
      * arbitrary values. The naming style for groups and keys is CamelCase.
      */
-    QSettings *settings() const { return mSettings; }
+    QSettings *settings() const;
 
 public slots:
     void setShowGrid(bool showGrid);
     void setShowTileObjectOutlines(bool enabled);
     void setShowTileAnimations(bool enabled);
+    void setShowTileCollisionShapes(bool enabled);
     void setSnapToGrid(bool snapToGrid);
     void setSnapToFineGrid(bool snapToFineGrid);
     void setSnapToPixels(bool snapToPixels);
@@ -168,6 +189,7 @@ public slots:
     void setGridFine(int gridFine);
     void setObjectLineWidth(qreal lineWidth);
     void setHighlightCurrentLayer(bool highlight);
+    void setHighlightHoveredObject(bool highlight);
     void setShowTilesetGrid(bool showTilesetGrid);
     void setAutomappingDrawing(bool enabled);
     void setOpenLastFilesOnStartup(bool load);
@@ -180,6 +202,7 @@ signals:
     void showGridChanged(bool showGrid);
     void showTileObjectOutlinesChanged(bool enabled);
     void showTileAnimationsChanged(bool enabled);
+    void showTileCollisionShapesChanged(bool enabled);
     void snapToGridChanged(bool snapToGrid);
     void snapToFineGridChanged(bool snapToFineGrid);
     void snapToPixelsChanged(bool snapToPixels);
@@ -187,8 +210,10 @@ signals:
     void gridFineChanged(int gridFine);
     void objectLineWidthChanged(qreal lineWidth);
     void highlightCurrentLayerChanged(bool highlight);
+    void highlightHoveredObjectChanged(bool highlight);
     void showTilesetGridChanged(bool showTilesetGrid);
     void objectLabelVisibilityChanged(ObjectLabelVisiblity);
+    void labelForHoveredObjectChanged(bool enabled);
 
     void applicationStyleChanged(ApplicationStyle);
     void baseColorChanged(const QColor &baseColor);
@@ -208,11 +233,12 @@ signals:
 
     void recentFilesChanged();
 
-    void checkForUpdatesChanged();
+    void checkForUpdatesChanged(bool on);
+    void displayNewsChanged(bool on);
 
 private:
     Preferences();
-    ~Preferences();
+    ~Preferences() override;
 
     bool boolValue(const char *key, bool def = false) const;
     QColor colorValue(const char *key, const QColor &def = QColor()) const;
@@ -225,6 +251,7 @@ private:
     bool mShowGrid;
     bool mShowTileObjectOutlines;
     bool mShowTileAnimations;
+    bool mShowTileCollisionShapes;
     bool mSnapToGrid;
     bool mSnapToFineGrid;
     bool mSnapToPixels;
@@ -232,9 +259,11 @@ private:
     int mGridFine;
     qreal mObjectLineWidth;
     bool mHighlightCurrentLayer;
+    bool mHighlightHoveredObject;
     bool mShowTilesetGrid;
     bool mOpenLastFilesOnStartup;
     ObjectLabelVisiblity mObjectLabelVisibility;
+    bool mLabelForHoveredObject;
     ApplicationStyle mApplicationStyle;
     QColor mBaseColor;
     QColor mSelectionColor;
@@ -243,6 +272,8 @@ private:
     Map::RenderOrder mMapRenderOrder;
     bool mDtdEnabled;
     bool mSafeSavingEnabled;
+    bool mExportOnSave;
+    ExportOptions mExportOptions;
     QString mLanguage;
     bool mReloadTilesetsOnChange;
     bool mUseOpenGL;
@@ -255,16 +286,92 @@ private:
     QString mObjectTypesFile;
 
     QDate mFirstRun;
-    QDate mPatreonDialogTime;
+    QDate mDonationDialogTime;
     int mRunCount;
     bool mIsPatron;
     bool mCheckForUpdates;
+    bool mDisplayNews;
     bool mWheelZoomsByDefault;
 
     static Preferences *mInstance;
 };
 
 
+inline bool Preferences::showGrid() const
+{
+    return mShowGrid;
+}
+
+inline bool Preferences::showTileObjectOutlines() const
+{
+    return mShowTileObjectOutlines;
+}
+
+inline bool Preferences::showTileAnimations() const
+{
+    return mShowTileAnimations;
+}
+
+inline bool Preferences::showTileCollisionShapes() const
+{
+    return mShowTileCollisionShapes;
+}
+
+inline bool Preferences::snapToGrid() const
+{
+    return mSnapToGrid;
+}
+
+inline bool Preferences::snapToFineGrid() const
+{
+    return mSnapToFineGrid;
+}
+
+inline bool Preferences::snapToPixels() const
+{
+    return mSnapToPixels;
+}
+
+inline QColor Preferences::gridColor() const
+{
+    return mGridColor;
+}
+
+inline int Preferences::gridFine() const
+{
+    return mGridFine;
+}
+
+inline qreal Preferences::objectLineWidth() const
+{
+    return mObjectLineWidth;
+}
+
+inline bool Preferences::highlightCurrentLayer() const
+{
+    return mHighlightCurrentLayer;
+}
+
+inline bool Preferences::highlightHoveredObject() const
+{
+    return mHighlightHoveredObject;
+}
+
+inline bool Preferences::showTilesetGrid() const
+{
+    return mShowTilesetGrid;
+}
+
+inline Preferences::ObjectLabelVisiblity Preferences::objectLabelVisibility() const
+{
+    return mObjectLabelVisibility;
+}
+
+inline bool Preferences::labelForHoveredObject() const
+{
+    return mLabelForHoveredObject;
+}
+
 inline Preferences::ApplicationStyle Preferences::applicationStyle() const
 {
     return mApplicationStyle;
@@ -280,14 +387,59 @@ inline QColor Preferences::selectionColor() const
     return mSelectionColor;
 }
 
+inline Map::LayerDataFormat Preferences::layerDataFormat() const
+{
+    return mLayerDataFormat;
+}
+
+inline Map::RenderOrder Preferences::mapRenderOrder() const
+{
+    return mMapRenderOrder;
+}
+
 inline bool Preferences::safeSavingEnabled() const
 {
     return mSafeSavingEnabled;
 }
 
-inline Preferences::ObjectLabelVisiblity Preferences::objectLabelVisibility() const
+inline bool Preferences::exportOnSave() const
 {
-    return mObjectLabelVisibility;
+    return mExportOnSave;
+}
+
+inline Preferences::ExportOptions Preferences::exportOptions() const
+{
+    return mExportOptions;
+}
+
+inline bool Preferences::exportOption(ExportOption option) const
+{
+    return mExportOptions.testFlag(option);
+}
+
+inline QString Preferences::language() const
+{
+    return mLanguage;
+}
+
+inline bool Preferences::reloadTilesetsOnChange() const
+{
+    return mReloadTilesetsOnChange;
+}
+
+inline bool Preferences::useOpenGL() const
+{
+    return mUseOpenGL;
+}
+
+inline bool Preferences::automappingDrawing() const
+{
+    return mAutoMapDrawing;
+}
+
+inline QString Preferences::mapsDirectory() const
+{
+    return mMapsDirectory;
 }
 
 inline QDate Preferences::firstRun() const
@@ -310,6 +462,11 @@ inline bool Preferences::checkForUpdates() const
     return mCheckForUpdates;
 }
 
+inline bool Preferences::displayNews() const
+{
+    return mDisplayNews;
+}
+
 inline bool Preferences::openLastFilesOnStartup() const
 {
     return mOpenLastFilesOnStartup;
@@ -320,5 +477,11 @@ inline bool Preferences::wheelZoomsByDefault() const
     return mWheelZoomsByDefault;
 }
 
-} // namespace Internal
+inline QSettings *Preferences::settings() const
+{
+    return mSettings;
+}
+
 } // namespace Tiled
+
+Q_DECLARE_OPERATORS_FOR_FLAGS(Tiled::Preferences::ExportOptions)
