@@ -30,22 +30,20 @@
 #include <QApplication>
 
 using namespace Tiled;
-using namespace Tiled::Internal;
 
 TileSelectionTool::TileSelectionTool(QObject *parent)
-    : AbstractTileTool(tr("Rectangular Select"),
-                       QIcon(QLatin1String(
-                               ":images/22x22/stock-tool-rect-select.png")),
-                       QKeySequence(tr("R")),
-                       parent)
-    , mSelectionMode(Replace)
+    : AbstractTileSelectionTool(tr("Rectangular Select"),
+                                QIcon(QLatin1String(
+                                      ":images/22/stock-tool-rect-select.png")),
+                                QKeySequence(Qt::Key_R),
+                                parent)
     , mMouseDown(false)
     , mSelecting(false)
 {
     setTilePositionMethod(OnTiles);
 }
 
-void TileSelectionTool::tilePositionChanged(const QPoint &)
+void TileSelectionTool::tilePositionChanged(QPoint)
 {
     if (mSelecting)
         brushItem()->setTileRegion(selectedArea());
@@ -84,23 +82,13 @@ void TileSelectionTool::mouseMoved(const QPointF &pos, Qt::KeyboardModifiers mod
 void TileSelectionTool::mousePressed(QGraphicsSceneMouseEvent *event)
 {
     const Qt::MouseButton button = event->button();
-    const Qt::KeyboardModifiers modifiers = event->modifiers();
 
     if (button == Qt::LeftButton) {
-        if (modifiers == Qt::ControlModifier) {
-            mSelectionMode = Subtract;
-        } else if (modifiers == Qt::ShiftModifier) {
-            mSelectionMode = Add;
-        } else if (modifiers == (Qt::ControlModifier | Qt::ShiftModifier)) {
-            mSelectionMode = Intersect;
-        } else {
-            mSelectionMode = Replace;
-        }
-
         mMouseDown = true;
         mMouseScreenStart = event->screenPos();
         mSelectionStart = tilePosition();
         brushItem()->setTileRegion(QRegion());
+        return;
     }
 
     if (button == Qt::RightButton) {
@@ -109,10 +97,14 @@ void TileSelectionTool::mousePressed(QGraphicsSceneMouseEvent *event)
             mSelecting = false;
             mMouseDown = false; // Avoid restarting select on move
             brushItem()->setTileRegion(QRegion());
-        } else {
+            return;
+        } else if (event->modifiers() == Qt::NoModifier) {
             clearSelection();
+            return;
         }
     }
+
+    AbstractTileTool::mousePressed(event);  // skipping AbstractTileSelection on purpose
 }
 
 void TileSelectionTool::mouseReleased(QGraphicsSceneMouseEvent *event)
@@ -127,7 +119,7 @@ void TileSelectionTool::mouseReleased(QGraphicsSceneMouseEvent *event)
         QRegion selection = document->selectedArea();
         const QRect area = selectedArea();
 
-        switch (mSelectionMode) {
+        switch (selectionMode()) {
         case Replace:   selection = area; break;
         case Add:       selection += area; break;
         case Subtract:  selection -= area; break;
@@ -152,7 +144,8 @@ void TileSelectionTool::mouseReleased(QGraphicsSceneMouseEvent *event)
 void TileSelectionTool::languageChanged()
 {
     setName(tr("Rectangular Select"));
-    setShortcut(QKeySequence(tr("R")));
+
+    AbstractTileSelectionTool::languageChanged();
 }
 
 QRect TileSelectionTool::selectedArea() const
