@@ -20,18 +20,16 @@
 
 #pragma once
 
+#include <QActionGroup>
 #include <QObject>
 
 class QAction;
-class QActionGroup;
 
 namespace Tiled {
 
 class Tile;
 class ObjectTemplate;
 
-namespace Internal {
-
 class AbstractTool;
 class MapDocument;
 
@@ -45,6 +43,7 @@ class MapDocument;
 class ToolManager : public QObject
 {
     Q_OBJECT
+    Q_DISABLE_COPY(ToolManager)
 
 public:
     ToolManager(QObject *parent = nullptr);
@@ -53,14 +52,23 @@ public:
     void setMapDocument(MapDocument *mapDocument);
 
     QAction *registerTool(AbstractTool *tool);
+    void unregisterTool(AbstractTool *tool);
 
-    void selectTool(AbstractTool *tool);
+    bool selectTool(AbstractTool *tool);
     AbstractTool *selectedTool() const;
 
+    template<typename Tool>
+    Tool *findTool();
+
+    QAction *findAction(AbstractTool *tool) const;
+
     void retranslateTools();
 
     void createShortcuts(QWidget *parent);
 
+    Tile *tile() const;
+    ObjectTemplate *objectTemplate() const;
+
 public slots:
     /**
      * Sets the tile that will be used when the creation mode is
@@ -80,12 +88,11 @@ signals:
 
 private slots:
     void actionTriggered(QAction *action);
+    void toolChanged();
     void toolEnabledChanged(bool enabled);
     void selectEnabledTool();
 
 private:
-    Q_DISABLE_COPY(ToolManager)
-
     AbstractTool *firstEnabledTool() const;
     void setSelectedTool(AbstractTool *tool);
 
@@ -100,6 +107,20 @@ private:
     bool mSelectEnabledToolPending;
 };
 
+/**
+ * Selects the tool that matches the specified type.
+ */
+template<class Tool>
+Tool *ToolManager::findTool()
+{
+    const auto actions = mActionGroup->actions();
+    for (QAction *action : actions) {
+        AbstractTool *abstractTool = action->data().value<AbstractTool*>();
+        if (Tool *tool = qobject_cast<Tool*>(abstractTool))
+            return tool;
+    }
+    return nullptr;
+}
 
 /**
  * Returns the selected tool.
@@ -109,5 +130,24 @@ inline AbstractTool *ToolManager::selectedTool() const
     return mSelectedTool;
 }
 
-} // namespace Internal
+inline Tile *ToolManager::tile() const
+{
+    return mTile;
+}
+
+inline ObjectTemplate *ToolManager::objectTemplate() const
+{
+    return mObjectTemplate;
+}
+
+inline void ToolManager::setTile(Tile *tile)
+{
+    mTile = tile;
+}
+
+inline void ToolManager::setObjectTemplate(ObjectTemplate *objectTemplate)
+{
+    mObjectTemplate = objectTemplate;
+}
+
 } // namespace Tiled
